# -*- coding: utf-8 -*-
"""
Provides configurations to be used in HTTP clients.
"""
import codecs
import json
import os
import os.path
from typing import Dict, Optional, Tuple

from systemlink.clientconfig.configuration import Configuration
from systemlink.messagebus.constants import Environment
from systemlink.messagebus.paths import (
    get_http_configurations_directory,
    get_legacy_http_configurations_directory,
    get_certificates_directory,
    get_legacy_certificates_directory)
from systemlink.messagebus.utils import is_windows, is_linux

HTTP_MASTER_CONFIGURATION_ID = 'SYSTEMLINK_MASTER'
HTTP_LOCALHOST_CONFIGURATION_ID = 'SYSTEMLINK_LOCALHOST'
HTTP_ENVIRONMENT_CONFIGURATION_ID = 'SYSTEMLINK_ENVIRONMENT'

_TEMP_CERT_PATH = '/run/systemlink/certs/http.crt'


def get_configuration(route_name: str) -> Configuration:
    """
    Gets a generic HTTP configuration for a service.

    :param route_name: The service's route name, e.g. 'nitestmonitor'
    :type route_name: str
    :return: A configuration for the service.
    :rtype: Configuration
    """
    return get_configuration_by_id(None, route_name, True)


def get_configuration_by_id(
        configuration_id: Optional[str],
        route_name: str,
        enable_fallbacks: bool
    ) -> Configuration:
    """
    Gets a specific HTTP configuration for a service.

    :param configuration_id: The ID of the configuration to look for,
        or None for default configurations.
    :type configuration_id: Optional[str]
    :param route_name: The service's route name.
    :type route_name: str
    :param enable_fallbacks: Whether or not to fallback to other known
        configurations if the requested configuration is not available.
    :type enable_fallbacks: bool
    :return: A configuration for the service.
    :rtype: Configuration
    """
    configurations = _read_configurations(route_name)
    if configuration_id is None:
        if enable_fallbacks:
            fallback_configuration = _fallback(configurations)
            if fallback_configuration is None:
                raise ValueError('No SystemLink configurations available.')
            return fallback_configuration
        raise ValueError('configuration_id cannot be null if enable_fallbacks is also false.')

    normalized_id = configuration_id.upper()
    selected_configuration = configurations.get(normalized_id)
    if selected_configuration is not None:
        return selected_configuration

    if not enable_fallbacks:
        raise ValueError('Configuration with ID {0} was not found.'.format(normalized_id))
    fallback_configuration = _fallback(configurations)
    if fallback_configuration is None:
        raise ValueError(
            'Configuration with ID {0} was not found and no fallback configurations were available.'
            .format(normalized_id))
    return fallback_configuration


def _fallback(configurations: Dict[str, Configuration]) -> Optional[Configuration]:
    """
    Performs fallback logic if configurations cannot be found.

    :param configurations: The dictionary of discovered configurations.
    :type configurations: Dict[str, Configuration]
    :return: The fallback configuration, or None if the fallback was not successful.
    :rtype: Optional[Configuration]
    """
    environment_configuration = configurations.get(HTTP_ENVIRONMENT_CONFIGURATION_ID)
    if environment_configuration is not None:
        return environment_configuration
    localhost_configuration = configurations.get(HTTP_LOCALHOST_CONFIGURATION_ID)
    if localhost_configuration is not None:
        return localhost_configuration
    master_configuration = configurations.get(HTTP_MASTER_CONFIGURATION_ID)
    return master_configuration


def _read_configurations(route_name: str) -> Dict[str, Configuration]:
    """
    Discovers and reads HTTP configurations.

    :param route_name: The route name to use when generating service configurations.
    :type route_name: str
    :return: A dict mapping each discovered configuration ID to its corresponding configuration.
    :rtype: Dict[str, Configuration]
    """
    configurations = {}  # type: Dict[str, Configuration]
    if is_linux():
        # Check to see if there are any files in the legacy Linux
        # location.
        search_path = get_legacy_http_configurations_directory()
        if os.path.isdir(search_path):
            _read_configuration_files(search_path, True, route_name, configurations)

    search_path = get_http_configurations_directory()
    if os.path.isdir(search_path):
        _read_configuration_files(search_path, False, route_name, configurations)
    env_config = _read_configuration_from_env(route_name)
    if env_config:
        configurations[HTTP_ENVIRONMENT_CONFIGURATION_ID] = env_config
    return configurations


def _read_configuration_from_env(route_name: str) -> Optional[Configuration]:
    """
    Attempt to read a configuration by reading the environment variables.

    :param route_name: The route name to use when generating service configurations.
    :type route_name: str
    :return: The configuration stored in environment variables, or None if
        no such configuration exists.
    :rtype: Optional[Configuration]
    """
    uri = os.environ.get(Environment.HTTP_URI)
    if not uri:
        return None
    api_key = os.environ.get(Environment.HTTP_API_KEY)
    cert = os.environ.get(Environment.HTTP_CERTIFICATE)
    cert_path = os.environ.get(Environment.HTTP_CERTIFICATE_PATH)

    # The API key and certificate are optional.
    actual_cert_path = None
    if cert_path:
        actual_cert_path = cert_path
    elif cert:
        if is_windows():
            raise NotImplementedError(
                'Passing a TLS certificate via the environment is not supported on Windows'
            )
        temp_cert_dir = os.path.dirname(_TEMP_CERT_PATH)
        if not os.path.isdir(temp_cert_dir):
            os.makedirs(temp_cert_dir)
        with open(_TEMP_CERT_PATH, 'w') as fp_:
            fp_.write(cert)
        actual_cert_path = _TEMP_CERT_PATH

    configuration_dict = {
        'Id': HTTP_ENVIRONMENT_CONFIGURATION_ID,
        'Uri': uri
    }
    if api_key:
        configuration_dict['ApiKey'] = api_key

    return Configuration(configuration_dict, route_name, actual_cert_path)


def _read_configuration_files(
        search_path: str,
        is_legacy_path: bool,
        route_name: str,
        configurations: Dict[str, Configuration]
    ) -> None:
    """
    Discovers and reads HTTP configurations.

    :param search_path: The path to use to search for files.
    :type search_path: str
    :param is_legacy_path: If ``True``, using a legacy path on Linux.
        ``False`` otherwise.
    :type is_legacy_path: bool
    :param route_name: The route name to use when generating service configurations.
    :type route_name: str
    :param configurations: A dict mapping each discovered configuration
        ID to its corresponding configuration. Used for both input and
        output.
    :type configurations: Dict[str, Configuration]
    """
    json_files = [os.path.join(search_path, name)
                  for name in os.listdir(search_path)
                  if name.endswith('.json')
                  and os.path.isfile(os.path.join(search_path, name))]
    for file_path in json_files:
        configuration_id, configuration = _read_config_file(file_path, is_legacy_path, route_name)
        if configuration_id is not None:
            configurations[configuration_id] = configuration


def _read_config_file(
        path: str,
        is_legacy_path: bool,
        route_name: str
    ) -> Tuple[Optional[str], Optional[Configuration]]:
    """
    Generates a configuration from a configuration file.

    :param path: The path of the configuration file.
    :type path: str
    :param is_legacy_path: If ``True``, using a legacy path on Linux.
        ``False`` otherwise.
    :type is_legacy_path: bool
    :param route_name: The route name to use when generating a service configuration.
    :type route_name: str
    :return: A 2-tuple containing the configuration ID and the configuration,
        or (None, None) if the configuration could not be read.
    :rtype: Tuple[Optional[str], Optional[Configuration]]
    """
    if not os.path.isfile(path):
        return None, None

    with codecs.open(path, 'r') as file:
        configuration_dict = json.loads(file.read())
    configuration_id = configuration_dict.get('Id')
    if configuration_id is not None:
        cert_path = configuration_dict.get('CertPath')
        if cert_path is not None:
            cert_path = _get_certificate_path(configuration_dict['CertPath'], is_legacy_path)
        configuration_id = configuration_dict['Id'].upper()
        configuration = Configuration(configuration_dict, route_name, cert_path)
        return configuration_id, configuration
    return None, None


def _get_certificate_path(cert_path: str, is_legacy_path: bool) -> str:
    """
    Gets the full path to a certificate.

    :param cert_path: The configured certificate path.
    :type cert_path: str
    :return: The absolute path of the certificate.
    :rtype: str
    """
    if os.path.isabs(cert_path):
        return cert_path
    if is_legacy_path:
        path = os.path.join(
            get_legacy_certificates_directory(),
            cert_path
        )
    else:
        path = os.path.join(
            get_certificates_directory(),
            cert_path
        )
    return path
