# -*- coding: utf-8 -*-
"""
Implementation of 'ApiKeyAuthHandler'.
"""
from __future__ import annotations

# Import python libs
import logging
import threading
from datetime import datetime, timedelta
from typing import Dict, List, Tuple

# Import local libs
from systemlink.messagebus.auth.auth_result import AuthResult
from systemlink.messagebus.auth.i_auth_handler import IAuthHandler
from systemlink.messagebus.auth.user_services_client import UserServicesClient

# Set up logging
LOGGER = logging.getLogger(__name__)


class ApiKeyAuthHandler(IAuthHandler):  # pylint: disable=too-few-public-methods
    """
    Auth handler which authenticates the caller based on the given
    API key and returns user and policy information so that services
    can authorize the caller.
    """
    _cache_lock: threading.Lock = threading.Lock()
    _cache: Dict[str, Tuple[AuthResult, datetime]] = {}

    def __init__(self) -> None:
        self._user_services_client: UserServicesClient = UserServicesClient()

    def auth(self, token: str) -> AuthResult:
        """
        The Auth method takes a token (e.g. apiKey) and authenticates
        the caller.

        :param token: The token to authenticate
        :type token: str
        :return: The AuthResult indicates whether or not the
           authentication of the caller was successful.
        :rtype: AuthResult
        """
        with self._cache_lock:
            if self._cache:
                self._perform_cache_expiration()
                cache_item = self._cache.get(token)
                if cache_item is not None:
                    return cache_item[0]

        try:
            auth_result = self._user_services_client.auth(token)
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.error('Authentication error: %s', exc, exc_info=True)
            auth_result = AuthResult.create_failure(str(exc))
        with self._cache_lock:
            expiry_timestamp = datetime.utcnow() + timedelta(seconds=5)
            self._cache[token] = (auth_result, expiry_timestamp)

        return auth_result

    def _perform_cache_expiration(self) -> None:
        """
        Remove expired tokens from authenication cache
        """
        now = datetime.utcnow()
        remove_list: List[str] = []
        for token, (_, expiry_timestamp) in self._cache.items():
            if now > expiry_timestamp:
                remove_list.append(token)
        for token in remove_list:
            del self._cache[token]
