# -*- coding: utf-8 -*-
"""
Implementation of 'AuthPolicy'.
"""
# Import python libs
from typing import cast, Any, Dict, List

# Import local libs
from systemlink.messagebus.auth.auth_statement import AuthStatement


class AuthPolicy:
    """
    A policy contains a list of statements which describe what a user
    is allowed to do in the system.

    It defines all actions (operations) the user is allowed to call
    and which resources the user is allowed to access.

    Policies are always additive which means a policy cannot deny access but only grant it.

    This is a minimal implementation of the UserService AuthPolicy class that can be used
    for /auth requests
    """
    def __init__(
            self,
            statements: List[AuthStatement]
        ) -> None:
        """
        :param statements: The statements of the policy.
        :type statements: List[AuthStatement]
        """
        self._statements = statements

    @property
    def statements(self) -> List[AuthStatement]:
        """
        Get the statements.

        :return: The statements.
        :rtype: List[AuthStatement]
        """
        return self._statements

    def serialize(self) -> Dict[str, List[Any]]:
        """
        Serialize the policy to JSON.

        :return: The policy, as JSON.
        :rtype: Dict[str, List[Any]]
        """
        return {
            'statements': [statement.serialize() for statement in self._statements]
        }

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, AuthPolicy):
            return False

        return self.statements == cast(AuthPolicy, other).statements

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return 'AuthPolicy(statements={0})'.format(self.statements)

    __repr__ = __str__


ADMIN_POLICY_INFO = AuthPolicy([AuthStatement(['*'], ['*'], '*')])
