# -*- coding: utf-8 -*-
"""
Implementation of 'AuthPolicyTemplate'.
"""
# Import python libs
from typing import cast, Any, List

# Import local libs
from systemlink.messagebus.auth.auth_statement import AuthStatement


class AuthPolicyTemplate:
    """
    A policy template defines what a user or device is allowed
    to do but is not bound to a workspace.

    When the template is referenced by a policy, the connection
    between the actions and the workspace is created.
    """
    def __init__(
            self,
            id_: str,
            statements: List[AuthStatement]
        ) -> None:
        """
        :param id_: The id of the policy template
        :type id_: str
        :param statements: The statements of the policy.
        :type statements: List[AuthStatement]
        """
        self._id = id_
        self._statements = statements

    @property
    def id(self) -> str: # pylint: disable=invalid-name
        """
        Get the id.

        :return: The id.
        :rtype: str
        """
        return self._id

    @property
    def statements(self) -> List[AuthStatement]:
        """
        Get the statements.

        :return: The statements.
        :rtype: List[AuthStatement]
        """
        return self._statements

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, AuthPolicyTemplate):
            return False

        return (
            self.id == cast(AuthPolicyTemplate, other).id and
            self.statements == cast(AuthPolicyTemplate, other).statements
        )

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return 'AuthPolicyTemplate(id={0}, statements={1})'.format(self.id, self.statements)

    __repr__ = __str__
