# -*- coding: utf-8 -*-
"""
Implementation of 'AuthResult'.
"""
# Import python libs
from typing import cast, Any, List, Optional

# Import local libs
from systemlink.messagebus.auth.auth_org import AuthOrg
from systemlink.messagebus.auth.auth_policy import AuthPolicy
from systemlink.messagebus.auth.auth_status import AuthStatus
from systemlink.messagebus.auth.auth_user import AuthUser
from systemlink.messagebus.auth.auth_workspace import AuthWorkspace


class AuthResult:
    """
    Result structure which holds all the authentication information
    about the caller.

    If the authentication was successful, the structure contains user
    and policy information so that services can perform authorization.

    If the authentication failed, the structure contains the reason
    for the failure.
    """
    def __init__(  # pylint: disable=too-many-arguments
            self,
            status: AuthStatus,
            user: Optional[AuthUser],
            org: Optional[AuthOrg],
            workspaces: List[AuthWorkspace],
            policies: List[AuthPolicy],
            failure_reason: Optional[str]
        ) -> None:
        """
        :param status: The auth status.
        :type status: AuthStatus
        :param user: The user information.
        :type user: Optional[AuthUser]
        :param org: The org information.
        :type org: Optional[AuthOrg]
        :param workspaces: The workspaces.
        :type workspaces: List[AuthWorkspace]
        :param policies: the policies for authorizing users' access to
           resources.
        :type policies: List[AuthPolicy]
        :param failure_reason: The reason for the authentication failure.
        :type failure_reason: Optional[str]
        """
        self._status = status
        self._user = user
        self._org = org
        self._workspaces = workspaces
        self._policies = policies
        self._failure_reason = failure_reason

    @classmethod
    def create_empty(cls) -> 'AuthResult':
        """
        Creates a new instance of AuthResult which indicates a
        successful authentication but without any other information
        since there were no actual credentials.

        Should be used as migration for insecure access for AMQP
        messages that do not yet have APIKeys.

        :return: The successful auth result.
        :rtype: AuthResult
        """
        return cls(AuthStatus.SUCCESS, None, None, [], [], None)

    @classmethod
    def create_success(
            cls,
            user: AuthUser,
            org: AuthOrg,
            workspaces: List[AuthWorkspace],
            policies: List[AuthPolicy]
        ) -> 'AuthResult':
        """
        Creates a new instance of AuthResult which indicates a
        successful authentication.

        :param user: The user information.
        :type user: AuthUser
        :param org: The org information.
        :type org: AuthOrg
        :param workspaces: The workspaces.
        :type workspaces: List[AuthWorkspace]
        :param policies: the policies for authorizing users' access to
           resources.
        :return: The successful auth result.
        :rtype: AuthResult
        """
        return cls(AuthStatus.SUCCESS, user, org, workspaces, policies, None)

    @classmethod
    def create_failure(
            cls,
            failure_reason: str
        ) -> 'AuthResult':
        """
        Creates a new instance of AuthResult which indicates a failed
        authentication.

        :param failure_reason: The reason for the authentication failure.
        :type failure_reason: str
        :return: The auth result marked as failed.
        :rtype: AuthResult
        """
        return cls(AuthStatus.FAILURE, None, None, [], [], failure_reason)

    @property
    def status(self) -> AuthStatus:
        """
        Get the auth status.

        :return: The auth status.
        :rtype: AuthStatus
        """
        return self._status

    @property
    def user(self) -> Optional[AuthUser]:
        """
        Get the user information.

        :return: The user information.
        :rtype: Optional[AuthUser]
        """
        return self._user

    @property
    def org(self) -> Optional[AuthOrg]:
        """
        Get the org information.

        :return: The org information.
        :rtype: Optional[AuthOrg]
        """
        return self._org

    @property
    def workspaces(self) -> List[AuthWorkspace]:
        """
        Get the workspaces.

        :return: The workspaces.
        :rtype: List[AuthWorkspace]
        """
        return self._workspaces

    @property
    def policies(self) -> List[AuthPolicy]:
        """
        Get the policies for authorizing users' access to resources.

        :return: The policies for authorizing users' access to
           resources.
        :rtype: List[AuthPolicy]
        """
        return self._policies

    @property
    def failure_reason(self) -> Optional[str]:
        """
        Get the reason for the authentication failure.

        :return: The reason for the authentication failure.
        :rtype: Optional[str]
        """
        return self._failure_reason

    @property
    def default_workspace(self) -> Optional[AuthWorkspace]:
        """
        Get the default workspace.

        :return: The default workspace or ``None`` is not found.
        :rtype: Optional[AuthWorkspace]
        """
        for workspace in self._workspaces:
            if workspace.is_default:
                return workspace
        return None

    def get_workspace(self, id_: str) -> Optional[AuthWorkspace]:
        """
        Get the workspace based on workspace ID.

        :param id_: The workspace ID.
        :type id_: str
        :return: The workspace or ``None`` is not found.
        :rtype: Optional[AuthWorkspace]
        """
        for workspace in self._workspaces:
            if workspace.id == id_:
                return workspace
        return None

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, AuthResult):
            return False

        return (
            self.status == cast(AuthResult, other).status and
            self.user == cast(AuthResult, other).user and
            self.org == cast(AuthResult, other).org and
            self.workspaces == cast(AuthResult, other).workspaces and
            self.policies == cast(AuthResult, other).policies and
            self.failure_reason == cast(AuthResult, other).failure_reason
        )

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return (
            'AuthResult(status={0}, user={1}, org={2}, workspaces={3}, '
            'policies={4}, failure_reason={5})'.format(
                self.status, self.user, self.org, self.workspaces,
                self.policies, self.failure_reason
            )
        )

    __repr__ = __str__
