# -*- coding: utf-8 -*-
"""
Implementation of 'AuthUser'.
"""
# Import python libs
from typing import cast, Any, Optional


class AuthUser:
    """
    The user information of the caller.

    Initially, SLS will only support a single organization.
    """
    def __init__(  # pylint: disable=too-many-arguments
            self,
            id_: str,
            org_id: str,
            login: str,
            first_name: Optional[str],
            last_name: Optional[str],
            email: Optional[str]
        ) -> None:
        """
        :param id_: The user id.
        :type id_: str
        :param org_id: The organization id.
        :type org_id: str
        :param login: The login.
        :type login: str
        :param first_name: The first name.
        :type first_name: Optional[str]
        :param last_name: The last name.
        :type last_name: Optional[str]
        :param email: The email.
        :type email: Optional[str]
        """
        self._id = id_
        self._org_id = org_id
        self._login = login
        self._first_name = first_name
        self._last_name = last_name
        self._email = email

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """
        Get the user ID.

        :return: The user ID.
        :rtype: str
        """
        return self._id

    @property
    def org_id(self) -> str:
        """
        Get the organization ID.

        :return: The organization ID.
        :rtype: str
        """
        return self._org_id

    @property
    def login(self) -> str:
        """
        Get the login.

        :return: The login.
        :rtype: str
        """
        return self._login

    @property
    def first_name(self) -> Optional[str]:
        """
        Get the first name.

        :return: The first name.
        :rtype: Optional[str]
        """
        return self._first_name

    @property
    def last_name(self) -> Optional[str]:
        """
        Get the last name.

        :return: The last name.
        :rtype: Optional[str]
        """
        return self._last_name

    @property
    def email(self) -> Optional[str]:
        """
        Get the email.

        :return: The email.
        :rtype: Optional[str]
        """
        return self._email


    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, AuthUser):
            return False

        return (
            self.id == cast(AuthUser, other).id and
            self.org_id == cast(AuthUser, other).org_id and
            self.login == cast(AuthUser, other).login and
            self.first_name == cast(AuthUser, other).first_name and
            self.last_name == cast(AuthUser, other).last_name and
            self.email == cast(AuthUser, other).email
        )

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return (
            'AuthUser(id={0}, org_id={1}, login={2}, first_name={3}, '
            'last_name={4}, email={5})'.format(
                self.id, self.org_id, self.login, self.first_name,
                self.last_name, self.email
            )
        )

    __repr__ = __str__
