# -*- coding: utf-8 -*-
"""
Implementation of 'PolicyDefinition'.
"""
# Import python libs
from typing import cast, Any, Dict, List

# Import local libs
from systemlink.messagebus.auth.statement_definition import StatementDefinition


class PolicyDefinition:
    """
    A policy contains a list of statements which describe what a user
    is allowed to do in the system.

    It defines all actions (operations) the user is allowed to call
    and which resources the user is allowed to access.

    Policies are always additive which means a policy cannot deny access but only grant it.

    This class should be 1-1 with the policy definition defined in the auth service
    swagger document.
    """
    def __init__(
            self,
            id_: str,
            statements: List[StatementDefinition]
        ) -> None:
        """
        :param id_: The id of the policy
        :type id_: str
        :param statements: The statements of the policy.
        :type statements: List[StatementDefinition]
        """
        self._id = id_
        self._statements = statements

    @property
    def id(self) -> str: # pylint: disable=invalid-name
        """
        Get the id.

        :return: The id.
        :rtype: str
        """
        return self._id

    @property
    def statements(self) -> List[StatementDefinition]:
        """
        Get the statements.

        :return: The statements.
        :rtype: List[StatementDefinition]
        """
        return self._statements

    def serialize(self) -> Dict[str, List[Any]]:
        """
        Serialize the policy to JSON.

        :return: The policy, as JSON.
        :rtype: Dict[str, List[Any]]
        """
        return {
            'statements': [statement.serialize() for statement in self._statements]
        }

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, PolicyDefinition):
            return False

        return (
            self.id == cast(PolicyDefinition, other).id and
            self.statements == cast(PolicyDefinition, other).statements
        )

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return 'PolicyDefinition(id={0}, statements={1})'.format(self.id, self.statements)

    __repr__ = __str__


ADMIN_POLICY_INFO = PolicyDefinition(None, [StatementDefinition(['*'], ['*'], '*', '*')])
