# -*- coding: utf-8 -*-
"""
Implementation of 'PolicyEvaluator'.
"""
from __future__ import annotations

# Import python libs
import re
from typing import Iterable, List, Set, Tuple

# Import local libs
from systemlink.messagebus.auth.auth_policy import AuthPolicy
from systemlink.messagebus.auth.auth_statement import AuthStatement
from systemlink.messagebus.auth.resource_id import ResourceId


class PolicyEvaluator:
    """
    The policy evaluator takes a list of policies and evaluates whether
    or not the given resourceIds and requiredAction actually match any
    of the statements.

    Policies and statements are additive which means if only one
    statement matches the permission will be granted.

    Statements contain a list of allowed resource identifiers and
    action names. Resources, actions and workspace support * as a
    wildcard for zero or more characters.

    e.g. AuthStatement(new[] { "file:Delete" }, new[] { "FileA" },
         "WorkspaceA")
    This statements grants the user access to delete the file with the
    name 'FileA' in workspace 'WorkspaceA'.

    AuthStatement(new[] { "file:*" }, new[] { "*.txt" }, "*")
    This statements grants the user access to perform any operation
    (create, delete, download, etc...) on text files in all workspaces.
    """
    @classmethod
    def get_resource_pattern(
            cls,
            required_action: str,
            policies: Iterable[AuthPolicy]
        ) -> List[ResourceId]:
        """
        This evaluates all policies and returns the resource and
        workspace pattern for the given action. The action is matched
        supporting the wildcard *.

        :param required_action: The action to find resources for.
        :type required_action: str
        :param policies: The policies.
        :type policies: Iterable[AuthPolicy]
        :return: A list of resource ID patterns.
        :rtype: List[ResourceId]
        """
        result: Set[ResourceId] = set()
        for policy in policies:
            cls._resource_ids_by_policy(result, required_action, policy)
        return list(result)

    @classmethod
    def filter_resources(
            cls,
            resource_ids: Iterable[ResourceId],
            required_action: str,
            policies: Iterable[AuthPolicy]
        ) -> List[ResourceId]:
        """
        Returns all resource identifiers from the list which are
        allowed for the required action.

        :param resource_ids: The resource identifiers to check.
        :type resource_ids: Iterable[ResourceId]
        :param required_action: The action to find resources for.
        :type required_action: str
        :param policies: The policies.
        :type policies: Iterable[AuthPolicy]
        :return: A list of allowed resource identifiers.
        :rtype: List[ResourceId]
        """
        authorized: List[ResourceId] = []
        for resource_id in resource_ids:
            if cls.is_resource_allowed(resource_id, required_action, policies):
                authorized.append(resource_id)
        return authorized

    @classmethod
    def split_resources(
            cls,
            resource_ids: Iterable[ResourceId],
            required_action: str,
            policies: Iterable[AuthPolicy]
        ) -> Tuple[List[ResourceId], List[ResourceId]]:
        """
        Returns all resource identifiers split into two lists: the id
        list of the authorized resources and the id list of the
        unauthorized resources for the required action.

        :param resource_ids: The resource identifiers to check.
        :type resource_ids: Iterable[ResourceId]
        :param required_action: The action to find resources for.
        :type required_action: str
        :param policies: The policies.
        :type policies: Iterable[AuthPolicy]
        :return: Tuple of two lists: a list of allowed policy
            identifiers, a list of not allowed policy indetifiers.
        :rtype: Tuple[List[ResourceId], List[ResourceId]]
        """
        authorized: List[ResourceId] = []
        unauthorized: List[ResourceId] = []
        for resource_id in resource_ids:
            if cls.is_resource_allowed(resource_id, required_action, policies):
                authorized.append(resource_id)
            else:
                unauthorized.append(resource_id)
        return authorized, unauthorized

    @classmethod
    def has_action(
            cls,
            required_action: str,
            policies: Iterable[AuthPolicy]
        ) -> bool:
        """
        Evaluates if the policies allow performing the required action
        on at least one workspace.

        :param required_action: The action to filter.
        :type required_action: str
        :param policies: The policies.
        :type policies: Iterable[AuthPolicy]
        :return: True, if the policy contains the action at least once.
            False, if there are no possible resource-workspace combinations
            for which the action would be allowed.
        """
        for policy in policies:
            if cls._has_action(required_action, policy):
                return True
        return False

    @classmethod
    def _has_action(
            cls,
            required_action: str,
            statement: AuthStatement
        ) -> bool:
        """
        Evaluates if the given auth statement allows performing the
        required action.

        :param required_action: The action to filter.
        :type required_action: str
        :param statement: The auth statement to check.
        :type statement: AuthStatement
        :return: True, if the given auth statement allows performing
            the required action.
        """
        return cls._any_match(required_action, statement.actions)

    @classmethod
    def is_allowed(
            cls,
            resource_ids: Iterable[ResourceId],
            required_action: str,
            policies: Iterable[AuthPolicy]
        ) -> bool:
        """
        Evaluates if the policies allow access to all given resource
        identifiers for the required action.

        :param resource_ids: The resource identifiers to check.
        :type resource_ids: Iterable[ResourceId]
        :param required_action: The action to filter.
        :type required_action: str
        :param policies: The policies.
        :type policies: Iterable[AuthPolicy]
        :return: ``True`` if access is allowed to all given resource
            identifiers. ``False`` otherwise.
        :rtype: bool
        """
        for resource_id in resource_ids:
            if not cls.is_resource_allowed(resource_id, required_action, policies):
                return False
        return True

    @classmethod
    def is_resource_allowed(
            cls,
            resource_id: ResourceId,
            required_action: str,
            policies: Iterable[AuthPolicy]
        ) -> bool:
        """
        Evaluates if the policies allow access to the given resource
        identifier for the required action.

        :param resource_id: The resource identifier to check.
        :type resource_id: ResourceId
        :param required_action: The action to filter.
        :type required_action: str
        :param policies: The policies.
        :type policies: Iterable[AuthPolicy]
        :return: ``True`` if access is allowed to the given resource
            identifier. ``False`` otherwise.
        :rtype: bool
        """
        for policy in policies:
            if cls._is_allowed_by_policy(resource_id, required_action, policy):
                return True
        return False

    @classmethod
    def _is_allowed_by_policy(
            cls,
            resource_id: ResourceId,
            required_action: str,
            policy: AuthPolicy
        ) -> bool:
        """
        Evaluates if the policy allows access to the given resource
        identifier for the required action.

        :param resource_id: The resource identifier to check.
        :type resource_id: ResourceId
        :param required_action: The action to filter.
        :type required_action: str
        :param policy: The policy.
        :type policy: AuthPolicy
        :return: ``True`` if access is allowed to the given resource
            identifier. ``False`` otherwise.
        :rtype: bool
        """
        for statement in policy.statements:
            if cls._is_allowed_by_statement(resource_id, required_action, statement):
                return True
        return False

    @classmethod
    def _is_allowed_by_statement(
            cls,
            resource_id: ResourceId,
            required_action: str,
            statement: AuthStatement
        ) -> bool:
        """
        Evaluates if the statement allows access to the given resource
        identifier for the required action.

        :param resource_id: The resource identifier to check.
        :type resource_id: ResourceId
        :param required_action: The action to filter.
        :type required_action: str
        :param statement: The statement.
        :type statement: AuthStatement
        :return: ``True`` if access is allowed to the given resource
            identifier. ``False`` otherwise.
        :rtype: bool
        """
        return (
            cls._any_match(required_action, statement.actions) and
            cls._any_match(resource_id.resource, statement.resources) and
            cls._is_allowed_workspace(resource_id, statement)
        )

    @classmethod
    def _is_allowed_workspace(
            cls,
            resource_id: ResourceId,
            statement: AuthStatement
        ) -> bool:
        """
        Evaluates if the statement allows access to the workspace
        associated with the given resource identifier.

        :param resource_id: The resource identifier to check.
        :type resource_id: ResourceId
        :param statement: The statement.
        :type statement: AuthStatement
        :return: ``True`` if access is allowed to the workspace.
            ``False`` otherwise.
        :rtype: bool
        """
        if not resource_id.has_workspace:
            return True
        return cls._matches(resource_id.workspace, statement.workspace)

    @classmethod
    def _resource_ids_by_policy(
            cls,
            result: Set[ResourceId],
            required_action: str,
            policy: AuthPolicy
        ) -> None:
        """
        Collects matching resource IDs by policy.

        :param result: The collection of matching resource IDs.
        :type result: Set[ResourceId]
        :param required_action: The action to filter.
        :type required_action: str
        :param policy: The policy.
        :type policy: AuthPolicy
        """
        for statement in policy.statements:
            cls._resource_ids_by_statement(result, required_action, statement)

    @classmethod
    def _resource_ids_by_statement(
            cls,
            result: Set[ResourceId],
            required_action: str,
            statement: AuthStatement
        ) -> None:
        """
        Collects matching resource IDs by statement.

        :param result: The collection of matching resource IDs.
        :type result: Set[ResourceId]
        :param required_action: The action to filter.
        :type required_action: str
        :param statement: The statement.
        :type statement: AuthStatement
        """
        if not cls._any_match(required_action, statement.actions):
            return

        for resource in statement.resources:
            result.add(ResourceId(resource, statement.workspace))

    @classmethod
    def _any_match(
            cls,
            match_str: str,
            filters: Iterable[str]
        ) -> bool:
        """
        Checks if any of the filters match.

        :param match_str: The string to match against.
        :type match_str: str
        :param filters: The filters to use.
        :type filters: Iterable[str]
        :return: ``True`` if any filters match the match string.
            ``False`` otherwise.
        :rtype: bool
        """
        for filter_ in filters:
            if cls._matches(match_str, filter_):
                return True
        return False


    @classmethod
    def _matches(
            cls,
            match_str: str,
            filter_: str
        ) -> bool:
        """
        Checks if the filter matches.

        :param match_str: The string to match against.
        :type match_str: str
        :param filter_: The filter to use.
        :type filter_: str
        :return: ``True`` if the filter matches the match string.
            ``False`` otherwise.
        :rtype: bool
        """
        pattern = cls._convert_to_pattern(filter_)
        return re.fullmatch(pattern, match_str) is not None

    @staticmethod
    def _convert_to_pattern(
            filter_: str
        ) -> str:
        """
        Converts a filter to a pattern.

        :param filter_: The filter to convert.
        :type filter_: str
        :return: The converted filter as a pattern.
        :rtype: str
        """
        escaped = re.escape(filter_)
        # No need to prefix with ``^`` and postfix with ``$``
        # because we will use :func:`re.fullmatch`.
        return escaped.replace('\\*', '.*')
