# -*- coding: utf-8 -*-
"""
Implementation of 'PolicyResource'.
"""
# Import python libs
from typing import cast, Any, Dict, List, Optional

# Import local libs
from systemlink.messagebus.auth.auth_policy import AuthPolicy
from systemlink.messagebus.auth.auth_statement import AuthStatement
from systemlink.messagebus.auth.policy_type import PolicyType


class PolicyResource:
    """
    A group of statements which describe what a user is allowed
    to do in the system.
    """
    def __init__(
            self,
            name: Optional[str],
            type_: PolicyType,
            statements: List[AuthStatement],
            properties: Optional[Dict[str, str]] = None
        ) -> None:
        """
        :param name: The name of the policy.
        :type name: Optional[str]
        :param type_: The intent of the policy.
        :type type_: PolicyType
        :param statements: The statements attached to the policy.
        :type statements: List[AuthStatement]
        :param properties: Key-value metadata associated with this policy.
        :type properties: Optional[Dict[str, str]]
        """
        self._name = name
        self._type = type_
        self._statements = statements
        self._properties = properties if properties else {}

    @classmethod
    def from_auth_policy(
            cls,
            policy: AuthPolicy,
            name: Optional[str] = None,
            type_: PolicyType = PolicyType.Internal,
            properties: Optional[Dict[str, str]] = None
        ):
        """
        Initializes a policy resource from a simplified policy and metadata.

        :param policy: The simplified policy.
        :type policy: AuthPolicy
        :param name: The name of the policy.
        :type name: Optional[str]
        :param type_: The intent of the policy.
        :type type_: PolicyType
        :param properties: Key-value metadata associated with this policy.
        :type properties: Optional[Dict[str, str]]
        :return: The constructed policy.
        :rtype: PolicyResource
        """
        return PolicyResource(
            name,
            type_,
            policy.statements,
            properties
        )

    @property
    def name(self) -> Optional[str]:
        """
        The name of this policy.

        :return: The name of this policy.
        :rtype: Optional[str]
        """
        return self._name

    @property
    def type_(self) -> PolicyType:
        """
        A value which indicates the intended use of this policy.

        :return: A value which indicates the intended use of this policy.
        :rtype: PolicyType
        """
        return self._type

    @property
    def statements(self) -> List[AuthStatement]:
        """
        Get the statements.

        :return: The statements.
        :rtype: List[AuthStatement]
        """
        return self._statements

    @property
    def properties(self) -> Dict[str, str]:
        """
        Key-value metadata associated with this policy.

        :return: Key-value metadata associated with this policy.
        :rtype: Dict[str, str]
        """
        return self._properties

    def serialize(self) -> Dict[str, Any]:
        """
        Serialize the policy to JSON.

        :return: The policy, as JSON.
        :rtype: Dict[str, List[Any]]
        """
        return {
            'name': self._name,
            'type': str(self.type_) if self.type_ else None,
            'statements': [statement.serialize() for statement in self._statements],
            'properties': self._properties
        }

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, PolicyResource):
            return False

        other = cast(PolicyResource, other)
        return self.statements == other.statements \
            and self.name == other.name \
            and self.type_ == other.type_ \
            and self.properties == other.properties

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return 'PolicyResource(name={0}, type={1}, statements={2}, properties={3})'.format(
            self.name,
            self.type_,
            self.statements,
            self.properties
        )

    __repr__ = __str__
