# -*- coding: utf-8 -*-
"""
Implementation of 'ResourceId'.
"""
from __future__ import annotations

# Import python libs
from typing import cast, Any, Optional


class ResourceId:
    """
    Resource identifier which typically consists of two parts:
    The resource name and the workspace the resource is part of.
    """
    def __init__(
            self,
            resource: str = '*',
            workspace: str = '*'
        ) -> None:
        """
        :param resource: The resource.
        :type resource: str
        :param workspace: The workspace.
        :type workspace: str
        """
        if resource is None:
            raise ValueError('Resource cannot be None')
        if workspace is None or workspace.isspace():
            raise ValueError('Workspace cannot be None or whitespace')
        self._resource: str = resource
        self._workspace: Optional[str] = workspace

    @property
    def has_workspace(self) -> bool:
        """
        Whether or not the resource is part of a workspace.

        :return: ``True`` if the resource is part of a workspace.
            ``False`` otherwise.
        :rtype: bool
        """
        return self._workspace is not None

    @property
    def resource(self) -> str:  # pylint: disable=invalid-name
        """
        Get the resource.

        :return: The resource.
        :rtype: str
        """
        return self._resource

    @property
    def workspace(self) -> Optional[str]:
        """
        Get the workspace.

        :return: The workspace.
        :rtype: Optional[str]
        """
        return self._workspace

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal.

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, ResourceId):
            return False

        return (
            self.resource == cast(ResourceId, other).resource and
            self.workspace == cast(ResourceId, other).workspace
        )

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal.

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __hash__(self) -> int:
        """
        Returns a hash for object comparisons.

        :return: A hash for object comparisons.
        :rtype: int
        """
        return hash((self.resource, self.workspace))

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return 'ResourceId(resource={0}, workspace={1})'.format(self.resource, self.workspace)

    __repr__ = __str__
