# -*- coding: utf-8 -*-
"""
Implementation of 'StatementDefinition'.
"""
# Import python libs
from typing import cast, Any, Dict, List


class StatementDefinition:
    """
    A statement defines which actions can be performed and which
    resources can be accessed in a workspace. This should be 1-1
    with the statement definition defined in the auth service
    swagger document
    """
    def __init__(
            self,
            actions: List[str],
            resources: List[str],
            workspace: str,
            type_: str
        ) -> None:
        """
        :param actions: A list of action patterns.
        :type actions: List[str]
        :param resources: A list of resource patterns.
        :type resources: List[str]
        :param workspace: A workspace pattern.
        :type workspace: str
        :param type_: The type of the resources these actions apply.
        :type type_: str
        """
        self._actions = actions
        self._resources = resources
        self._type = type_
        self._workspace = workspace

    @property
    def actions(self) -> List[str]:
        """
        Get the actions.

        :return: The actions.
        :rtype: List[str]
        """
        return self._actions

    @property
    def resources(self) -> List[str]:
        """
        Get the resources.

        :return: The resources.
        :rtype: List[str]
        """
        return self._resources

    @property
    def type(self) -> str:
        """
        Get the type of the resources these actions apply.

        :return: The type of the resources these actions apply.
        :rtype: str
        """
        return self._type

    @property
    def workspace(self) -> str:
        """
        Get the workspace.

        :return: The workspace.
        :rtype: str
        """
        return self._workspace

    def serialize(self) -> Dict[str, Any]:
        """
        Serialize the statement to JSON.

        :return: The statement, as JSON.
        :rtype: Dict[str, Any]
        """
        return {
            'actions': self._actions,
            'resource': self._resources,
            'type': self._type,
            'workspace': self._workspace
        }

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether both objects are equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are equal. ``False``
            otherwise.
        :rtype: bool
        """
        if not isinstance(other, StatementDefinition):
            return False

        return (
            self.actions == cast(StatementDefinition, other).actions and
            self.resources == cast(StatementDefinition, other).resources and
            self.type == cast(StatementDefinition, other).type and
            self.workspace == cast(StatementDefinition, other).workspace
        )

    def __ne__(self, other: Any) -> bool:
        """
        Returns whether both objects are not equal

        :param other: Other object to compare with.
        :type other: Any
        :return: ``True`` if both objects are not equal. ``False``
            otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self) -> str:
        """
        Returns a string representation of this object.

        :return: A string representation of this object.
        :rtype: str
        """
        return (
            'StatementDefinition(actions={0}, resources={1}, type={2}, workspace={3})'.format(
                self.actions, self.resources, self.type, self.workspace
            )
        )

    __repr__ = __str__
