# -*- coding: utf-8 -*-  # pylint: disable=too-many-lines
"""
Implementation of 'UserServicesClient'.
"""

# Import python libs
from http import HTTPStatus
import json
import os
from typing import cast, Any, Dict, List, Optional, Tuple

# Import third party libs
from urllib.parse import urlparse
import urllib3

# Import local libs
from systemlink.messagebus.auth.auth_key import AuthKey
from systemlink.messagebus.auth.auth_org import AuthOrg
from systemlink.messagebus.auth.auth_policy import AuthPolicy
from systemlink.messagebus.auth.auth_policy_template import AuthPolicyTemplate
from systemlink.messagebus.auth.auth_result import AuthResult
from systemlink.messagebus.auth.auth_statement import AuthStatement
from systemlink.messagebus.auth.auth_user import AuthUser
from systemlink.messagebus.auth.auth_workspace import AuthWorkspace
from systemlink.messagebus.auth.policy_definition import PolicyDefinition
from systemlink.messagebus.auth.policy_resource import PolicyResource
from systemlink.messagebus.auth.policy_type import PolicyType
from systemlink.messagebus.auth.statement_definition import StatementDefinition
from systemlink.messagebus.constants import Environment
from systemlink.messagebus.exceptions import SystemLinkException
from systemlink.messagebus.message_header import JSON_MESSAGE_CONTENT_TYPE

_DEFAULT_USER_SERVICES_URI: str = 'http://127.0.0.1:12100'
_GET_AUTH_URI: str = '/niauth/v1/auth'
_CREATE_SESSION_KEYS_ROUTE: str = '/niauth/v1/session-keys'
_SYSTEMLINK_SERVER_ORG_ROUTE: str = '/niuser/v1/orgs/SystemLink%20Server'
_KEYS_URI: str = '/niauth/v1/keys'
_POLICY_TEMPLATES_URI: str = '/niauth/v1/policy-templates'
_POLICIES_URI: str = '/niauth/v1/policies'
_API_KEY_HEADER: str = 'x-ni-api-key'
_ACCEPT_HEADER: str = 'Accept'
_CONTENT_TYPE_HEADER: str = 'Content-Type'


class UserServicesClient:
    """
    Http Client Wrapper for the User Services.
    Used to authenticate calls and retrieve user and policy
    information.
    """
    _default_workspace: Optional[AuthWorkspace] = None
    _org_id: str = None

    def __init__(self, user_services_uri: Optional[str] = None) -> None:
        if not user_services_uri:
            user_services_uri = self._get_user_services_uri()
        if not user_services_uri:
            user_services_uri = _DEFAULT_USER_SERVICES_URI
        self._user_services_uri = user_services_uri
        self._http_client: urllib3.PoolManager = urllib3.PoolManager()

    @staticmethod
    def _get_user_services_uri() -> Optional[str]:
        """
        Reads the user services URI from the environment.

        :return: The user services URI, or None, if the value
            is not present in the environment.
        :rtype: str
        """
        env_uri = os.getenv(Environment.USER_SERVICES_URI)
        if not env_uri or env_uri.isspace():
            return None
        parsed_url = urlparse(env_uri)
        if not parsed_url.scheme or not parsed_url.netloc:
            msg = 'User Services URL is invalid'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        return env_uri

    def get_default_workspace(self, whitelisted_api_key: Optional[str] = None) -> AuthWorkspace:
        """
        Get the default workspace.

        :param whitelisted_api_key: An API key to authenticate the request.
        :type whitelisted_api_key: Optional[str]
        :return: The default workspace.
        :rtype: AuthWorkspace
        """
        if self._default_workspace is None:
            self._retrieve_default_org_properties(whitelisted_api_key)
        return self._default_workspace

    def _retrieve_default_org_properties(self, whitelisted_api_key: Optional[str]) -> None:
        """
        Acquires default organization values such as the SystemLink Server
        org ID and default workspace.

        :param whitelisted_api_key: An API key to authenticate the request.
        :type whitelisted_api_key: Optional[str]
        :rtype: None
        """
        http_client = urllib3.PoolManager()
        uri = self._user_services_uri + _SYSTEMLINK_SERVER_ORG_ROUTE
        headers: Dict[str, str] = {
            _ACCEPT_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }
        # Eventually all calls should send a whitelisted API key.
        if whitelisted_api_key:
            headers[_API_KEY_HEADER] = whitelisted_api_key
        try:
            response = http_client.request(
                'GET', uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when asking UserService for default workspace: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status != HTTPStatus.OK.value:
            msg = (
                'Error when asking UserService for default workspace, Status Code: {0}'
                ''.format(response.status)
            )
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        self._set_org_id(json_dict)
        self._set_default_workspace(json_dict)

    @classmethod
    def _set_org_id(cls, response: Dict[str, Any]) -> None:
        """
        Sets the cached org ID from a Get Org response.

        :param response: The Get Org response to cache data from.
        :type response: Dict[str, Any]
        """
        org_id = response.get('id')
        if org_id is None or org_id == '':
            msg = 'Error when asking UserService for SystemLink Server Org Id: Org Id not found'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        cls._org_id = org_id

    @classmethod
    def _set_default_workspace(cls, response: Dict[str, Any]) -> None:
        """
        Sets the cached default workspace from a Get Org response.

        :param response: The Get Org response to cache data from.
        :type response: Dict[str, Any]
        """
        workspaces = cast(List[Dict[str, Any]], response.get('workspaces', []))
        for workspace_dict in workspaces:
            is_default = cast(bool, workspace_dict.get('default', False))
            if is_default:
                cls._default_workspace = cls._convert_to_workspace(workspace_dict)
                return
        msg = 'Error when asking for default workspace: Default workspace not found'
        raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

    def get_session_key(
            self,
            whitelisted_api_key: str,
            policies: List[AuthPolicy],
            duration_seconds: int = None
        ) -> str:
        """
        Creates a new API key for a single session.

        :param whitelisted_api_key: An API key which can create session keys.
        :type whitelisted_api_key: str
        :param policies: The policies associated with the new session key.
        :type policies: List[AuthPolicy]
        :param duration_seconds: The duration for which the session key will be valid.
            Must not exceed 24 hours.
        :type duration: int
        :return: The newly created session key.
        :rtype: str
        """
        if whitelisted_api_key is None or whitelisted_api_key.isspace():
            msg = 'Cannot send create session key request: Invalid whitelisted api key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not policies:
            msg = 'Cannot send create session key request: Invalid policies'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        (request, headers) = self._get_create_session_key_request(
            whitelisted_api_key,
            policies,
            None,
            None,
            duration_seconds
        )
        return self._send_create_session_key_request(request, headers)

    def get_session_key_by_user(
            self,
            whitelisted_api_key: str,
            user_id: str,
            duration_seconds: int = None
        ) -> str:
        """
        Creates a new API key for a single session, based on an existing user.

        :param whitelisted_api_key: An API key which can create session keys.
        :type whitelisted_api_key: str
        :param user_id: The ID of the user for which to create the session key.
        :type user_id: str
        :param duration_seconds: The duration for which the session key will be valid.
            Must not exceed 24 hours.
        :type duration_seconds: int
        :return: The newly created session key.
        :rtype: str
        """
        if whitelisted_api_key is None or whitelisted_api_key.isspace():
            msg = 'Cannot send create session key request: Invalid whitelisted api key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if user_id is None or user_id.isspace():
            msg = 'Cannot send create session key request: Invalid user ID'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        (request, headers) = self._get_create_session_key_request(
            whitelisted_api_key,
            None,
            None,
            user_id,
            duration_seconds
        )
        return self._send_create_session_key_request(request, headers)

    def get_session_key_by_policy_ids(
            self,
            whitelisted_api_key: str,
            policy_ids: List[str],
            duration_seconds: int = None
        ) -> str:
        """
        Creates a session key based on a list of pre-existing policies.

        :param whitelisted_api_key: An API key which can create session keys.
        :type whitelisted_api_key: str
        :param policy_ids: The IDs of the pre-existing policies.
        :type policy_ids: List[str]
        :param duration_seconds: The duration for which the session key will be valid.
            Must not exceed 24 hours.
        :type duration_seconds: int
        :return: The newly created session key.
        :rtype: str
        """
        if whitelisted_api_key is None or whitelisted_api_key.isspace():
            msg = 'Cannot send create session key request: Invalid whitelisted api key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not policy_ids:
            msg = 'Cannot send create session key request: Invalid policy IDs'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        (request, headers) = self._get_create_session_key_request(
            whitelisted_api_key,
            None,
            policy_ids,
            None,
            duration_seconds
        )
        return self._send_create_session_key_request(request, headers)

    def get_session_key_by_policies_and_user(
            self,
            whitelisted_api_key: str,
            policies: List[AuthPolicy],
            user_id: str,
            duration_seconds: int = None
        ) -> str:
        """
        Creates a session key based on a list of policies which is associated
        with the given user.

        :param whitelisted_api_key: An API key which can create session keys.
        :type whitelisted_api_key: str
        :param policies: The policies associated with the new session key.
        :type policies: List[AuthPolicy]
        :param user_id: The ID of the user for which to create the session key.
        :type user_id: str
        :param duration_seconds: The duration for which the session key will be valid.
            Must not exceed 24 hours.
        :type duration: int
        :return: The newly created session key.
        :rtype: str
        """
        if whitelisted_api_key is None or whitelisted_api_key.isspace():
            msg = 'Cannot send create session key request: Invalid whitelisted api key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not policies:
            msg = 'Cannot send create session key request: Invalid policies'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if user_id is None or user_id.isspace():
            msg = 'Cannot send create session key request: Invalid user ID'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        (request, headers) = self._get_create_session_key_request(
            whitelisted_api_key,
            policies,
            None,
            user_id,
            duration_seconds
        )
        return self._send_create_session_key_request(request, headers)

    def _get_create_session_key_request(  # pylint: disable=too-many-arguments
            self,
            whitelisted_api_key: str,
            policies: Optional[List[AuthPolicy]],
            policy_ids: Optional[List[str]],
            user_id: Optional[str],
            duration_seconds: int
        ) -> Tuple[Dict[str, Any], Dict[str, str]]:
        """
        Constructs the request to create a session key.

        :param whitelisted_api_key: An API key which can create session keys.
        :type whitelisted_api_key: str
        :param policies: The policies associated with the new session key.
        :type policies: Optional[List[AuthPolicy]]
        :param policy_ids: The IDs of the policies associated with the new session key.
        :type policy_ids: Optional[List[str]]
        :param user_id: The ID of the user for which to create the session key.
        :type user_id: Optional[str]
        :param duration: The duration for which the session key will be valid.
            Must not exceed 24 hours.
        :type duration:
        :return: A tuple containing the request body and request headers to send.
        :rtype: Tuple[Dict[str, Any], Dict[str, str]]
        """
        if duration_seconds is not None and duration_seconds > 24 * 60 * 60:
            raise ValueError("Duration cannot exceed 24 hours.")
        org_id = self._get_org_id(whitelisted_api_key)
        headers: Dict[str, str] = {
            _API_KEY_HEADER: whitelisted_api_key,
            _CONTENT_TYPE_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }
        body: Dict[str, Any] = {
            'orgId': org_id
        }
        if policies is not None:
            body['policies'] = [policy.serialize() for policy in policies]
        if policy_ids:
            body['policyIds'] = policy_ids
        if user_id is not None:
            body['userId'] = user_id
        if duration_seconds is not None:
            body['durationSeconds'] = duration_seconds
        return (json.dumps(body).encode('utf-8'), headers)

    def _send_create_session_key_request(
            self,
            body: Dict[str, Any],
            headers: Dict[str, str]
        ) -> str:
        """
        Sends the request to create a session key.

        :param body: The request body.
        :type body: Dict[str, Any]
        :param headers: The request headers.
        :type headers: Dict[str, str]
        :return: The newly created session key.
        :rtype: str
        """
        uri = self._user_services_uri + _CREATE_SESSION_KEYS_ROUTE
        try:
            response = self._http_client.request(
                'POST', uri, body=body, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending create session key request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status != HTTPStatus.OK.value:
            msg = (
                """Error when sending create session key request to UserService,
                Status Code: {0}, Response: {1}""".format(response.status, response.data)
            )
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        api_key = json_dict.get('secret')
        if api_key is None or api_key == '':
            msg = 'Error when sending create session key request: Did not create session key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        return api_key

    def _get_org_id(self, whitelisted_api_key: str) -> str:
        """
        Gets the ID of the SystemLink Server org.

        :param whitelisted_api_key: An API key to authenticate the request.
        :type whitelisted_api_key: str
        :return: The ID of the SystemLink Server org.
        :rtype: str
        """
        if self._org_id is None:
            self._retrieve_default_org_properties(whitelisted_api_key)
        return self._org_id

    def get_policy_templates(self, api_key, type_, name_, built_in_) -> List[AuthPolicyTemplate]:
        """
        Returns the list of policy-templates filtered by template type
        :param api_key: An api key with rights to query policy templates.
        :type api_key: str
        :param type_ The policy template type, can be omitted. Possible values are: user and service
        :type type_ str
        :param name_ The policy template name, can be omitted.
        :type name_ str
        :param built_in_ Whether the policy template is built in, can be omitted.
            Possible values are: 'true' or 'false'
        :type built_in_ str
        :return: a list with policy templates that match the given type
        """
        request_uri = self._user_services_uri + _POLICY_TEMPLATES_URI + '?'
        if type_ is not None:
            request_uri += 'type=' + type_ + '&'
        if name_ is not None:
            request_uri += 'name=' + name_ + '&'
        if built_in_ is not None:
            request_uri += 'builtIn=' + built_in_

        try:
            headers: Dict[str, str] = {
                _API_KEY_HEADER: api_key,
                _ACCEPT_HEADER: JSON_MESSAGE_CONTENT_TYPE
            }
            response = self._http_client.request(
                'GET', request_uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending request to UserService for Policy Templates: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status != HTTPStatus.OK.value:
            msg = """Error when asking UserService for Policy Templates,
            Status Code: {0}, Response: {1}""".format(response.status, response.data)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)
        return self._get_policy_template_from_response(response)

    def _get_policy_template_from_response(self, response) -> List[AuthPolicyTemplate]:
        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        if not json_dict['policyTemplates']:
            return None
        return self._convert_to_policy_templates(
            cast(List[Dict[str, Any]], json_dict['policyTemplates'])
        )

    def get_policy(self, api_key: str, id_: str) -> PolicyResource:
        """
        Gets a policy by ID.
        """
        if not api_key or api_key.isspace():
            msg = 'Cannot get policy: invalid API key.'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not id_ or id_.isspace():
            msg = 'Cannot get policy: invalid policy ID.'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _POLICIES_URI + '/' + id_
        headers = {
            _API_KEY_HEADER: api_key
        }
        try:
            response = self._http_client.request(
                'GET', uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending get policy request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        if response.status == HTTPStatus.NOT_FOUND.value:
            msg = 'The policy was not found.'
            raise SystemLinkException.from_name('Skyline.NotFound', info=msg)
        if response.status < 200 or response.status >= 300:
            msg = 'Error when getting policy by ID, status code: {0}, response: {1}'.format(
                response.status, response.data)
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        return self._deserialize_policy_resource(json_dict)

    def create_policy( # pylint: disable=too-many-arguments
            self, api_key, name, template_id, workspace_id, type_
        ) -> AuthPolicy:
        """
        Creates a new policy based on a policy template and a workspace id
        :param api_key: An api key with rights to create policies.
        :type api_key: str
        :param name: The policy name.
        :type name: str
        :param template_id: The policy template id.
        :type template_id: str
        :param workspace_id: The workspace id the policy will be bound to.
        :type workspace_id: str
        :param type_: The type of the policies. Possible values are: default,
        internal, custom, role.
        :type type_: str
        :return: The policy
        :rtype: AuthPolicy
        """
        if not api_key:
            msg = 'Cannot send create policy request: Invalid session key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not name:
            msg = 'Cannot send create policy request: Invalid policy name'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not template_id:
            msg = 'Cannot send create policy request: Invalid policy template id'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not workspace_id:
            msg = 'Cannot send create policy request: Invalid workspace'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not type_:
            msg = 'Cannot send create policy request: Invalid type'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _POLICIES_URI
        headers, body = self._get_create_policy_request(
            api_key, name, template_id, workspace_id, type_
        )
        try:
            response = self._http_client.request(
                'POST', uri, headers=headers, body=body
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending create policy request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status != HTTPStatus.OK.value:
            msg = """Error when asking UserService to create policy,
            Status Code: {0}, Response: {1}""".format(response.status, response.data)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)
        return self._get_policy_from_response(response)

    def create_policy_resource(self, api_key: str, policy: PolicyResource) -> str:
        """
        Creates a policy from a policy resource.

        :param api_key: The API key used to authenticate the request.
        :type api_key: str
        :param policy: The policy to create.
        :type policy: PolicyResource
        :return: The ID of the created policy.
        :rtype: str
        """
        if not policy:
            raise ValueError('Policy cannot be None.')
        if not api_key or api_key.isspace():
            msg = 'Cannot create policy: invalid API key.'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _POLICIES_URI
        headers = {
            _API_KEY_HEADER: api_key,
            _CONTENT_TYPE_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }
        body = json.dumps(policy.serialize()).encode('utf-8')
        try:
            response = self._http_client.request(
                'POST', uri, headers=headers, body=body
            )
        except Exception as exc:
            msg = 'Error when sending create policy request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if response.status < 200 or response.status >= 300:
            msg = 'Error when creating policy, status code: {0}, response: {1}'.format(
                response.status, response.data)
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        return json_dict.get('id')

    def delete_policy(self, api_key: str, id_: str) -> None:
        """
        Deletes a policy.

        :param api_key: The API key used to authenticate the request.
        :type api_key: str
        :param id_: The ID of the policy to delete.
        :type id_: str
        """
        if not api_key or api_key.isspace():
            msg = 'Cannot delete policy: invalid API key.'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not id_ or id_.isspace():
            msg = 'Cannot delete policy: invalid policy ID.'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _POLICIES_URI + '/' + id_
        headers = {
            _API_KEY_HEADER: api_key
        }
        try:
            response = self._http_client.request(
                'DELETE', uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending delete policy request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if response.status < 200 or response.status >= 300:
            msg = 'Error when deleting policy, status code: {0}, response: {1}'.format(
                response.status, response.data)
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

    @staticmethod
    def _get_create_policy_request(
            api_key, name, template_id, workspace_id, type_
        ) -> Tuple[Dict[str, str], str]:
        headers: Dict[str, str] = {
            _API_KEY_HEADER: api_key,
            _CONTENT_TYPE_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }
        body: Dict[str, str] = {
            'name': name,
            'properties': {},
            'templateId': template_id,
            'workspace': workspace_id,
            'type': type_
        }
        return headers, json.dumps(body).encode('utf-8')

    def _get_policy_from_response(self, response) -> List[AuthPolicyTemplate]:
        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        if not json_dict:
            msg = 'Error when asking UserService to create policy: Did not create policy'
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)
        return self._convert_to_policy_definition(cast(Dict[str, Any], json_dict))

    def _deserialize_policy_resource(self, content: Dict[str, Any]) -> PolicyResource:
        """
        Deserializes a policy resource from JSON.

        :param content: The JSON.
        :type content: Dict[str, Any]
        :return: The deserialized policy.
        :rtype: PolicyResource
        """
        type_value = cast(str, content.get('type'))
        type_ = PolicyType.deserialize(type_value) if type_value else None
        statements_list = content.get('statements')
        if statements_list is not None:
            statements = self._convert_to_statement_definitions(
                cast(List[Dict[str, Any]], statements_list)
            )
        else:
            statements = []
        return PolicyResource(
            cast(str, content.get('name')),
            type_,
            statements,
            cast(Dict[str, str], content.get('properties'))
        )


    def create_permanent_api_key(self, session_key, name, policy_ids, workspace_id) -> str:
        """
        Creates a permanent api key with the rights provided by policy ids in given workspace
        :param session_key: An api key with rights to create keys.
        :type session_key: str
        :param name: The api key name.
        :type name: str
        :param policyIds: A list with policy ids.
        :type policyIds: List[str]
        :param workspace_id: The workspace id the api key will have access to.
        :type workspace_id: str
        :return: The policy
        :rtype: AuthPolicy
        """
        if not session_key:
            msg = 'Cannot send create key request: Invalid session key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not name:
            msg = 'Cannot send create key request: Invalid api key name'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not policy_ids:
            msg = 'Cannot send create key request: Invalid policies'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not workspace_id:
            msg = 'Cannot send create key request: Invalid workspace'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _KEYS_URI
        headers, body = self._get_create_api_key_request(
            session_key, name, policy_ids, workspace_id
        )
        try:
            response = self._http_client.request(
                'POST', uri, headers=headers, body=body
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending create key request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status != HTTPStatus.OK.value:
            msg = """Error when asking UserService to create key,
            Status Code: {0}, Response: {1}""".format(response.status, response.data)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        api_key = json_dict.get('secret')
        if api_key is None or api_key == '':
            msg = 'Error when sending create key request: Did not create key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        return api_key

    @staticmethod
    def _get_create_api_key_request(
            api_key, name, policy_ids, workspace_id
        ) -> Tuple[Dict[str, str], str]:
        headers: Dict[str, str] = {
            _API_KEY_HEADER: api_key,
            _CONTENT_TYPE_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }
        body: Dict[str, str] = {
            'name': name,
            'properties': {},
            'policyIds': policy_ids,
            'defaultWorkspace': workspace_id,
        }
        return headers, json.dumps(body).encode('utf-8')

    def get_permanent_api_key(self, session_key, name) -> AuthKey:
        """
        Gets the permanent api key with the given name if it exists.
        Raises SystemLinkException on failure.
        :param session_key: An api key with rights to query keys.
        :type session_key: str
        :param name: The api key name.
        :type name: str
        :return: api key definition
        :rtype: AuthKey
        """
        if not session_key:
            msg = 'Cannot send get key request: Invalid session key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not name:
            msg = 'Cannot send get key request: Invalid api key name'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _KEYS_URI + '/?name=' + name
        headers: Dict[str, str] = {
            _API_KEY_HEADER: session_key,
            _ACCEPT_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }

        try:
            response = self._http_client.request(
                'GET', uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending get key request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status != HTTPStatus.OK.value:
            msg = """Error when asking UserService to get key,
            Status Code: {0}, Response: {1}""".format(response.status, response.data)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        return self._get_api_key_from_response(response)

    def _get_api_key_from_response(self, response) -> AuthKey:
        content = cast(str, response.data)
        json_dict = cast(Dict[str, Any], json.loads(content))
        keys = cast(List[Dict[str, Any]], json_dict.get('keys'))
        if not keys:
            msg = 'Error when sending get key request: Could not get key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        api_key_dict = next(iter(keys), None)
        if not api_key_dict:
            msg = 'Error when sending get key request: Could not get key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        return self._convert_to_auth_key(api_key_dict)

    def delete_permanent_api_key(self, session_key, id_) -> None:
        """
        Deletes the permanent api key with the given id.
        Raises SystemLinkException on failure.
        :param session_key: An api key with rights to delete keys.
        :type session_key: str
        :param id_: The api key id.
        :type id_: str
        :return: None
        :rtype: None
        """
        if not session_key:
            msg = 'Cannot send delete key request: Invalid session key'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)
        if not id_:
            msg = 'Cannot send delete key request: Invalid api key id'
            raise SystemLinkException.from_name('Skyline.InternalServiceError', info=msg)

        uri = self._user_services_uri + _KEYS_URI + '/' + id_
        headers: Dict[str, str] = {
            _API_KEY_HEADER: session_key,
            _ACCEPT_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }

        try:
            response = self._http_client.request(
                'DELETE', uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending delete key request: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status < 200 or response.status >= 300:
            msg = """Error when asking UserService to delete key,
            Status Code: {0}, Response: {1}""".format(response.status, response.data)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

    def auth(self, api_key: str) -> AuthResult:
        """
        Authenticate the given apiKey by calling the /niauth/v1/auth
        method on the user service.

        :param api_key: The API key.
        :type api_key: str
        :return: The AuthResult indicates whether or not the
           authentication of the caller was successful.
        :rtype: AuthResult
        """
        return self._call_auth(api_key)

    def _call_auth(self, api_key: str) -> AuthResult:
        """
        Authenticate the given apiKey by calling the /niauth/v1/auth
        method on the user service.

        :param api_key: The API key.
        :type api_key: str
        :return: The AuthResult indicates whether or not the
           authentication of the caller was successful.
        :rtype: AuthResult
        """
        uri = self._user_services_uri + _GET_AUTH_URI
        headers: Dict[str, str] = {
            _API_KEY_HEADER: api_key,
            _ACCEPT_HEADER: JSON_MESSAGE_CONTENT_TYPE
        }
        try:
            response = self._http_client.request(
                'GET', uri, headers=headers
            )
        except Exception as exc:  # pylint: disable=broad-except
            msg = 'Error when sending authentication request for API key: {0}'.format(exc)
            raise SystemLinkException.from_name('Skyline.Exception', info=msg)

        if response.status == HTTPStatus.OK.value:
            return self._convert_to_auth_result(cast(str, response.data))
        if response.status == HTTPStatus.UNAUTHORIZED.value:
            return AuthResult.create_failure('API Key is not valid')
        msg = 'Error when authenticating API key, Status Code: {0}'.format(response.status)
        raise SystemLinkException.from_name('Skyline.Exception', info=msg)

    @classmethod
    def _convert_to_auth_result(cls, content: str) -> AuthResult:
        """
        Convert the response content to an AuthResult.

        :param content: The response content.
        :type content: str
        :return: The :class:`AuthResult` instance.
        :rtype: AuthResult
        """
        json_dict = cast(Dict[str, Any], json.loads(content))
        user = cls._convert_to_user(cast(Dict[str, Any], json_dict['user']))
        org = cls._convert_to_org(cast(Dict[str, Any], json_dict['org']))
        workspaces = cls._convert_to_workspaces(
            cast(List[Dict[str, Any]], json_dict['workspaces'])
        )
        policies = cls._convert_to_auth_policies(
            cast(List[Dict[str, Any]], json_dict['policies'])
        )
        return AuthResult.create_success(user, org, workspaces, policies)

    @staticmethod
    def _convert_to_user(user_dict: Dict[str, Any]) -> AuthUser:
        """
        Convert a dictionary to an AuthResult.

        :param user_dict: A dictionary representing an AuthUser.
        :type user_dict: Dict[str, Any]
        :return: The :class:`AuthUser` instance.
        :rtype: AuthUser
        """
        id_ = cast(str, user_dict['id'])
        org_id = cast(str, user_dict['orgId'])
        login = cast(str, user_dict['niuaId'])
        first_name = cast(Optional[str], user_dict.get('firstName'))
        last_name = cast(Optional[str], user_dict.get('lastName'))
        email = cast(Optional[str], user_dict.get('email'))
        return AuthUser(id_, org_id, login, first_name, last_name, email)

    @staticmethod
    def _convert_to_org(org_dict: Dict[str, Any]) -> AuthOrg:
        """
        Convert a dictionary to an AuthOrg.

        :param org_dict: A dictionary representing an AuthOrg.
        :type org_dict: Dict[str, Any]
        :return: The :class:`AuthOrg` instance.
        :rtype: AuthOrg
        """
        id_ = cast(str, org_dict['id'])
        name = cast(str, org_dict['name'])
        owner_id = cast(str, org_dict['ownerId'])
        return AuthOrg(id_, name, owner_id)

    @classmethod
    def _convert_to_workspaces(cls, workspaces: List[Dict[str, Any]]) -> List[AuthWorkspace]:
        """
        Convert a list of dictionaries to a list of AuthWorkspace.

        :param workspaces: A list of dictionaries representing
            workspaces.
        :type workspaces: List[Dict[str, Any]]
        :return: A list of :class:`AuthWorkspace` instances.
        :rtype: List[AuthWorkspace]
        """
        return [cls._convert_to_workspace(workspace) for workspace in workspaces]

    @staticmethod
    def _convert_to_workspace(workspace_dict: Dict[str, Any]) -> AuthWorkspace:
        """
        Convert a dictionary to an AuthWorkspace.

        :param workspace_dict: A dictionary representing an AuthWorkspace.
        :type workspace_dict: Dict[str, Any]
        :return: The :class:`AuthWorkspace` instance.
        :rtype: AuthWorkspace
        """
        id_ = cast(str, workspace_dict['id'])
        name = cast(str, workspace_dict['name'])
        enabled = cast(bool, workspace_dict.get('enabled', False))
        is_default = cast(bool, workspace_dict.get('default', False))
        return AuthWorkspace(id_, name, enabled, is_default)

    @classmethod
    def _convert_to_auth_policies(cls, policies: List[Dict[str, Any]]) -> List[AuthPolicy]:
        """
        Convert a list of dictionaries to a list of AuthPolicy.

        :param policies: A list of dictionaries representing
            policies.
        :type policies: List[Dict[str, Any]]
        :return: A list of :class:`AuthPolicy` instances.
        :rtype: List[AuthPolicy]
        """
        return [cls._convert_to_auth_policy(policy) for policy in policies]

    @classmethod
    def _convert_to_auth_policy(cls, policy_dict: Dict[str, Any]) -> AuthPolicy:
        """
        Convert a dictionary to an AuthPolicy.

        :param policy_dict: A dictionary representing an AuthPolicy.
        :type policy_dict: Dict[str, Any]
        :return: The :class:`AuthPolicy` instance.
        :rtype: AuthPolicy
        """
        statements = cls._convert_to_auth_statements(
            cast(List[Dict[str, Any]], policy_dict['statements'])
        )
        return AuthPolicy(statements)

    @classmethod
    def _convert_to_policiy_definitions(
            cls, policies: List[Dict[str, Any]]
        ) -> List[PolicyDefinition]:
        """
        Convert a list of dictionaries to a list of PolicyDefinition.

        :param policies: A list of dictionaries representing
            policies.
        :type policies: List[Dict[str, Any]]
        :return: A list of :class:`PolicyDefinition` instances.
        :rtype: List[PolicyDefinition]
        """
        return [cls._convert_to_policy_definition(policy) for policy in policies]

    @classmethod
    def _convert_to_policy_definition(cls, policy_dict: Dict[str, Any]) -> PolicyDefinition:
        """
        Convert a dictionary to an PolicyDefinition.

        :param policy_dict: A dictionary representing an PolicyDefinition.
        :type policy_dict: Dict[str, Any]
        :return: The :class:`PolicyDefinition` instance.
        :rtype: PolicyDefinition
        """
        id_ = cast(str, policy_dict.get('id', None))
        statements: List[AuthStatement] = []
        statements_list = policy_dict.get('statements', None)
        if statements_list is not None:
            statements = cls._convert_to_statement_definitions(
                cast(List[Dict[str, Any]], statements_list)
            )
        return PolicyDefinition(id_, statements)

    @classmethod
    def _convert_to_policy_templates(
            cls, policy_templates: List[Dict[str, Any]]
        ) -> List[AuthPolicyTemplate]:
        """
        Convert a list of dictionaries to a list of AuthPolicyTemplate.

        :param policy_templates: A list of dictionaries representing
            policy templates.
        :type policy_templates: List[Dict[str, Any]]
        :return: A list of :class:`AuthPolicyTemplate` instances.
        :rtype: List[AuthPolicyTemplate]
        """
        return [
            cls._convert_to_policy_template(policy_template)
            for policy_template in policy_templates
            ]

    @classmethod
    def _convert_to_policy_template(
            cls, policy_template_dict: Dict[str, Any]
        ) -> AuthPolicyTemplate:
        """
        Convert a dictionary to an AuthPolicyTemplate.

        :param policy_dict: A dictionary representing an AuthPolicyTemplate.
        :type policy_dict: Dict[str, Any]
        :return: The :class:`AuthPolicyTemplate` instance.
        :rtype: AuthPolicyTemplate
        """
        id_ = cast(str, policy_template_dict['id'])
        statements = cls._convert_to_statement_definitions(
            cast(List[Dict[str, Any]], policy_template_dict['statements'])
        )
        return AuthPolicyTemplate(id_, statements)

    @classmethod
    def _convert_to_auth_statements(cls, statements: List[Dict[str, Any]]) -> List[AuthStatement]:
        """
        Convert a list of dictionaries to a list of AuthStatement.

        :param policies: A list of dictionaries representing
            policies.
        :type policies: List[Dict[str, Any]]
        :return: A list of :class:`AuthStatement` instances.
        :rtype: List[AuthStatement]
        """
        return [cls._convert_to_auth_statement(statement) for statement in statements]

    @staticmethod
    def _convert_to_auth_statement(statement_dict: Dict[str, Any]) -> AuthStatement:
        """
        Convert a dictionary to an AuthStatement.

        :param statement_dict: A dictionary representing an AuthPolicy.
        :type statement_dict: Dict[str, Any]
        :return: The :class:`AuthStatement` instance.
        :rtype: AuthStatement
        """
        actions = cast(List[str], statement_dict['actions'])
        resources = cast(List[str], statement_dict['resource'])
        workspace = cast(str, statement_dict['workspace'])
        return AuthStatement(actions, resources, workspace)

    @classmethod
    def _convert_to_statement_definitions(
            cls, statements: List[Dict[str, Any]]
        ) -> List[StatementDefinition]:
        """
        Convert a list of dictionaries to a list of StatementDefinition.

        :param policies: A list of dictionaries representing
            policies.
        :type policies: List[Dict[str, Any]]
        :return: A list of :class:`StatementDefinition` instances.
        :rtype: List[StatementDefinition]
        """
        return [cls._convert_to_statement_definition(statement) for statement in statements]

    @staticmethod
    def _convert_to_statement_definition(statement_dict: Dict[str, Any]) -> StatementDefinition:
        """
        Convert a dictionary to an StatementDefinition.

        :param statement_dict: A dictionary representing an StatementDefinition.
        :type statement_dict: Dict[str, Any]
        :return: The :class:`StatementDefinition` instance.
        :rtype: StatementDefinition
        """
        actions = cast(List[str], statement_dict['actions'])
        resources = cast(List[str], statement_dict['resource'])
        workspace = cast(str, statement_dict.get('workspace', None))
        type_ = cast(str, statement_dict.get('type', None))
        return StatementDefinition(actions, resources, workspace, type_)

    @staticmethod
    def _convert_to_auth_key(statement_dict: Dict[str, Any]) -> AuthKey:
        """
        Convert a dictionary to an AuthKey.

        :param statement_dict: A dictionary representing an AuthKey.
        :type statement_dict: Dict[str, Any]
        :return: The :class:`AuthKey` instance.
        :rtype: AuthKey
        """
        id_ = cast(str, statement_dict.get('id', None))
        name = cast(str, statement_dict.get('name', None))
        return AuthKey(id_, name)
