# -*- coding: utf-8 -*-
"""
Collection of datetime-related functions.
"""
from __future__ import absolute_import

# Import python libs
import datetime
import re
from typing import Union

# Import third party libs
try:
    from dateutil import tz
    from dateutil.parser import isoparse
    HAS_DATEUTIL = True
except ImportError:
    HAS_DATEUTIL = False

TO_DATETIME_MS_INDEX = 6
TO_DATETIME_REGEX = re.compile(r'^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})\.?(\d{0,9})?Z?$')


def from_datetime(datetime_object: datetime.datetime) -> str:
    """
    Converts a :class:`datetime.datetime` object into a string
    which SystemLink interprets as a timestamp. The output
    string is in ISO 8601 format that ends with 'Z'.

    :param datetime_object: A datetime object representing a timestamp.
    :type datetime_object: datetime.datetime
    :return: A string that is in ISO 8601 format that ends with 'Z'.
    :rtype: str
    """
    # Assume that if there is no timezone information, it is UTC time
    # and not local time. Unfortunately, both
    # ``datetime.datetime.now()`` and ``datetime.datetime.utcnow()``
    # don't set timezone information. So we are assuming the client is
    # either using utcnow() or is explicitly adding in the timezone
    # information. Note that
    # ``datetime.datetime.now(datetime.timezone.utc)`` explicitly adds
    # in the timezone information and uses UTC time, but objects
    # with a UTC time zone do not equate to objects with no time zone.
    # Hence such objects may not be compared to utcnow() objects,
    # and utcnow() objects are preferred due to legacy usage.
    if datetime_object.tzinfo is None:
        return datetime_object.isoformat() + 'Z'
    if datetime_object.tzinfo == datetime.timezone.utc:
        return datetime_object.isoformat().replace('+00:00', 'Z')
    # ``datetime.timezone.utc`` != ``tz.tzutc()``, so we check for
    # both.
    if HAS_DATEUTIL and datetime_object.tzinfo == tz.tzutc():
        return datetime_object.isoformat().replace('+00:00', 'Z')
    return datetime_object.astimezone(
        tz=datetime.timezone.utc).isoformat().replace('+00:00', 'Z')


def to_datetime(datetime_string: Union[str, datetime.datetime]) -> datetime.datetime:
    """
    Converts a string which SystemLink interprets as a timestamp
    into a :class:`datetime.datetime` object.
    The input string is in ISO 8601 format that ends with 'Z'.
    Will pass through if the input is already an instance of
    :class:`datetime.datetime`.

    :param datetime_string: A string that is in ISO 8601 format that
        ends with 'Z'. Will pass through if this is already an instance
        of :class:`datetime.datetime`.
    :type datetime_string: str or datetime.datetime
    :return: A datetime object representing a timestamp.
    :rtype: datetime.datetime
    """
    if isinstance(datetime_string, datetime.datetime):
        return datetime_string

    no_tz_match = TO_DATETIME_REGEX.match(datetime_string)
    if no_tz_match:
        no_tz_match_groups = list(no_tz_match.groups())
        ms_match = no_tz_match_groups[TO_DATETIME_MS_INDEX]
        if ms_match:
            if len(ms_match) > 6:
                ms_match = ms_match[:6]
            no_tz_match_groups[TO_DATETIME_MS_INDEX] = int(ms_match) * pow(
                10, 6 - len(ms_match)
            )  # type: ignore
        else:
            no_tz_match_groups[TO_DATETIME_MS_INDEX] = 0  # type: ignore
        parsed_date = datetime.datetime(*map(int, no_tz_match_groups))  # type: ignore
        return parsed_date

    datetime_object = isoparse(datetime_string)
    if datetime_object.tzinfo is None:
        # Assume UTC and not local time when tzinfo is missing.
        # Do not add a time zone for backwards compatibility.
        return datetime_object
    # Remove time zone for backwards compatibility.
    # ``datetime.datetime`` objects with UTC time zone do not equate
    # to ``datetime.datetime`` objects with no time zone information.
    # So we standardize on no time zone information for comparison
    # purposes.
    if datetime_object.tzinfo in (datetime.timezone.utc, tz.tzutc()):
        return datetime_object.replace(tzinfo=None)
    return datetime_object.astimezone(tz=datetime.timezone.utc).replace(tzinfo=None)
