# -*- coding: utf-8 -*-
"""
Functions that assist in registering and using error names, codes, and
messages.
"""
# Import python libs
from typing import Any, List, Optional

# Import local libs
from systemlink.messagebus.error_code import ErrorCode

# Since systemlink.messagebus.exceptions imports this module, we cannot
# import that module at the top level to avoid a circular dependency.

ERROR_NAME_TO_CODE_MAP = {
    '': ErrorCode('', 'Unknown', 'An unknown error occurred.')
}
ERROR_NAME_TO_NUMERIC_CODE_MAP = {}
NUMERIC_CODE_TO_ERROR_NAME_MAP = {}


def is_registered() -> bool:
    """
    Check if the Error Code Registry has any registered entries.

    :return: ``True`` if any entries were registered. ``False`` otherwise.
    :rtype: bool
    """
    return len(ERROR_NAME_TO_CODE_MAP) > 1


def register(
        category: str,
        name: str,
        message: str,
        numeric_error_code: Optional[int] = None
    ) -> ErrorCode:
    """
    Register an error code.

    :param category: The category of the error.
    :type category: str
    :param name: The name of the error.
    :type name: str
    :param message: The error message.
    :type message: str
    :param numeric_error_code: The associated numeric error code.
    :type numeric_error_code: Optional[int]
    :return: The created
        :class:`systemlink.messagebus.error_code.ErrorCode` instance.
    :rtype: ErrorCode
    """
    error_code_obj = ErrorCode(category, name, message)
    qualified_name = str(error_code_obj)
    ERROR_NAME_TO_CODE_MAP[qualified_name] = error_code_obj
    if numeric_error_code is not None:
        ERROR_NAME_TO_NUMERIC_CODE_MAP[qualified_name] = numeric_error_code
        NUMERIC_CODE_TO_ERROR_NAME_MAP[numeric_error_code] = qualified_name
    return error_code_obj


def _unknown_error_code(qualified_name: str) -> ErrorCode:
    """
    Return an error code based on an unknown
    qualified name.

    :param qualified_name: The qualified name of the error. It is in
        the form ``[category].[name]``.
    :type qualified_name: str
    :return: A
        :class:`systemlink.messagebus.error_code.ErrorCode` instance.
    :rtype: ErrorCode
    """
    sep_idx = qualified_name.find('.')
    if sep_idx == -1:
        category = ''
        name = ''
    else:
        category = qualified_name[0:sep_idx]
        name = qualified_name[sep_idx+1:]
    error_code_obj = ErrorCode(category, name)
    return error_code_obj


def lookup(qualified_name: str) -> ErrorCode:
    """
    Look up an error code.

    :param qualified_name: The qualified name of the error. It is in
        the form ``[category].[name]``.
    :type qualified_name: str
    :return: A
        :class:`systemlink.messagebus.error_code.ErrorCode` instance.
    :rtype: ErrorCode
    """
    error_code_obj = ERROR_NAME_TO_CODE_MAP.get(qualified_name)
    if error_code_obj is not None:
        return error_code_obj
    return _unknown_error_code(qualified_name)


def format_message(uninserted_message: str, args: Optional[List[Any]] = None) -> str:
    """
    Format an error message.

    :param uninserted_message: The unformatted error message.
    :type uninserted_message: str
    :param args: A list of arguments (if needed) for the error message.
    :type args: Optional[List[Any]]
    :return: The formatted error message.
    :rtype: str
    """
    if args:
        return uninserted_message.format(*args)
    return uninserted_message


def get_error_message_by_name(qualified_name: str, args: Optional[List[Any]] = None) -> str:
    """
    Get an error message based on the error qualified name.

    :param qualified_name: The qualified name of the error. It is in
        the form ``[category].[name]``.
    :type qualified_name: str
    :param args: A list of arguments (if needed) for the error message.
    :type args: Optional[List[Any]]
    :return: The formatted error message.
    :rtype: str
    """
    error_code_obj = lookup(qualified_name)
    msg = error_code_obj.message
    return format_message(msg, args)


def get_error_message_by_numeric_code(  # pylint: disable=invalid-name
        numeric_error_code: int,
        args: Optional[List[Any]] = None
    ) -> str:
    """
    Get an error message based on the error name.

    :param numeric_error_code: The numeric error code.
    :type numeric_error_code: int
    :param args: A list of arguments (if needed) for the error message.
    :type args: Optional[List[Any]]
    :return: The formatted error message.
    :rtype: str
    """
    error_name = get_error_name_by_numeric_code(numeric_error_code)
    return format_message(lookup(error_name).message, args)


def get_numeric_error_code(qualified_name: str) -> int:
    """
    Get the numeric error code based on the qualified name.

    :param qualified_name: The qualified name of the error. It is in
        the form ``[category].[name]``.
    :type qualified_name: str
    :return: The numeric error code.
    :rtype: int
    """
    return ERROR_NAME_TO_NUMERIC_CODE_MAP.get(qualified_name, -251000)


def get_error_name_by_numeric_code(numeric_error_code: int) -> str:
    """
    Get the qualified name of the error based on the numeric error
    code.

    :param numeric_error_code: The numeric error code.
    :type numeric_error_code: int
    :return: The qualified name of the error. It is in
        the form ``[category].[name]``.
    :rtype: str
    """
    return NUMERIC_CODE_TO_ERROR_NAME_MAP.get(numeric_error_code, '')
