# -*- coding: utf-8 -*-
"""
Implementation of 'SystemLinkException'.
"""
# Import python libs
from typing import Any, List, Optional

# Import local libs
from systemlink.messagebus.error import Error
from systemlink.messagebus.error_code import ErrorCode
from systemlink.messagebus.error_code_registry import lookup


class SystemLinkException(Exception):
    """
    Base exception class: all Skyline-specific exceptions should subclass this.
    """
    def __init__(self, error: Optional[Error]) -> None:
        """
        :param error: An instance of :class`systemlink.messagebus.error.Error`.
            May be ``None``.
        :type error: Optional[Error]
        """
        if error is None:
            error_code_obj = lookup('Skyline.Exception')
            error = Error(error_code=error_code_obj)
        message = error.info
        super(SystemLinkException, self).__init__(message)
        self.message = message
        self.error = error

    @classmethod
    def from_name(
            cls,
            qualified_name: Optional[str],
            info: Optional[str] = None,
            args: Optional[List[Any]] = None
        ) -> 'SystemLinkException':
        """
        Create a SystemLinkException object based on the qualified name.

        :param qualified_name: The qualified name of the error. It is in
            the form ``[category].[name]``. May be ``None``.
        :type qualified_name: Optional[str]
        :param info: Error information string. May be ``None`` in the cases
            of no error or if this information is desired to be auto-generated
            based on ``qualified_name`` and ``args``.
        :type info: Optional[str]
        :param args: A list of arguments associated with the error. May be
            ``None`` if the associated error does not have any arguments or
            if ``info`` is not ``None``.
        :type args: Optional[List[Any]]
        :return: An instance of :class:`SystemLinkException`.
        :rtype: SystemLinkException
        """
        error_obj = Error.from_name(qualified_name, info=info, args=args)
        return cls(error=error_obj)

    @classmethod
    def from_error_code(
            cls,
            error_code: Optional[ErrorCode],
            info: Optional[str] = None,
            args: Optional[List[Any]] = None
        ) -> 'SystemLinkException':
        """
        Create a SystemLinkException object based on an ErrorCode object.

        :param error_code: An instance of
            :class`systemlink.messagebus.error_code.ErrorCode`. May be
            ``None``.
        :type error_code: Optional[ErrorCode]
        :param info: Error information string. May be ``None`` in the cases
            of no error or if this information is desired to be auto-generated
            based on ``error_code`` and ``args``.
        :type info: Optional[str]
        :param args: A list of arguments associated with the error. May be
            ``None`` if the associated error does not have any arguments or
            if ``info`` is not ``None``.
        :type args: Optional[List[Any]]
        :return: An instance of :class:`SystemLinkException`.
        :rtype: SystemLinkException
        """
        if error_code is None:
            error_code = lookup('Skyline.Exception')
        error_obj = Error(error_code=error_code, info=info, args=args)
        return cls(error=error_obj)

    @classmethod
    def from_exception(cls, exc: Exception, include_stack: bool = True) -> 'SystemLinkException':
        """
        Create a SystemLinkException object based on an exception.

        :param exc: An instance of a class derived from
            :class:`Exception`.
        :type exc: Exception
        :param include_stack: Whether to include stack information in
            the error message.
        :type include_stack: bool
        :return: An instance of :class:`SystemLinkException`.
        :rtype: SystemLinkException
        """
        if isinstance(exc, cls):
            return exc
        error_obj = Error.from_exception(exc, include_stack=include_stack)
        return cls(error=error_obj)

    def __repr__(self) -> str:
        ret = '[{0}] {1}'.format(self.error.name, self.message)
        return ret.strip()

    __str__ = __repr__
