# -*- coding: utf-8 -*-
"""
Implementation of 'GenericResponse' class
"""
# Import python libs
from typing import Optional

# Import local libs
from systemlink.messagebus.error import Error
from systemlink.messagebus.message_base import MessageBase
from systemlink.messagebus.message_header import MessageHeader, JSON_MESSAGE_CONTENT_TYPE
from systemlink.messagebus.response_message import ResponseMessage


class GenericResponse(ResponseMessage):
    """
    Generic Response.

    A response message that is genericized so that the body is represented as
    bytes.
    """
    def __init__(self, header: MessageHeader, body: Optional[bytes]) -> None:
        """
        :param header: The message header.
        :type header: MessageHeader
        :param body: The raw message body.
        :type body: Optional[bytes]
        """
        # This constructor is only invoked through
        # :func:`GenericResponse.from_routing_source`. It is never
        # invoked directly.
        self._request_name = ''
        super().__init__(
            header,
            body
        )

    @classmethod
    def from_routing_source(
            cls,
            routing_source: MessageBase,
            error: Optional[Error] = None
        ) -> 'GenericResponse':
        """
        Create a GenericResponse object based on routing source.

        :param routing_source: A message to use to identify
            the routing information for this response message.
        :type routing_source: MessageBase
        :param error: Optional error information to transfer to this response
            message.
        :type error: Optional[Error]
        :return: A :class:`GenericResponse` object with the routing
            information and optionally error information set.
        :rtype: GenericResponse
        """
        # Once we move to Python 3.7, we can use
        # `from __future__ import annotations`, and then use typing.cast
        # to cast to this class instead of marking things for mypy to ignore.
        instance = cls.from_message_name_content_type_routing_param(
            cls.__name__,
            JSON_MESSAGE_CONTENT_TYPE,
            routing_source.reply_to
        )
        # pylint: disable=protected-access
        instance._request_name = routing_source.message_name  # type: ignore
        instance.error = error
        instance.correlation_id = routing_source.correlation_id
        return instance  # type: ignore

    @property
    def request_name(self) -> str:
        """
        Get the request name.

        :return: The request name.
        :rtype: str
        """
        return self._request_name

    @property
    def message_body_as_bytes(self) -> bytes:
        """
        Get the message body in a raw bytes form.

        :return: The message body as bytes.
        :rtype: bytes
        """
        body = '{\"RequestName\":\"' + self._request_name + '\"}'
        return bytes(body, 'utf-8')
