# -*- coding: utf-8 -*-
"""
Implementation of 'LogBuffer' class
"""
# Import python libs
import datetime
from typing import List, Optional

# Import local libs
from systemlink.messagebus.trace_logger_messages import LogEntry, LogType
from systemlink.messagebus.trace_point import TracePoint


class LogBuffer(object):
    """
    Log Buffer.

    Buffer to be used for Trace Logger logs.
    """
    def __init__(self):
        self._capacity = 500
        self._entries = []

    def __len__(self) -> int:
        """
        Get the current number of entries in the log buffer.

        :return: The current number of entries in the log buffer.
        :rtype: int
        """
        return len(self._entries)

    def __bool__(self) -> bool:
        """
        Get whether the log buffer is not empty.

        :return: ``True`` if the log buffer is not empty. ``False``
            otherwise.
        :rtype: bool
        """
        return bool(self._entries)

    def reset(self) -> None:
        """
        Remove all entries from the log buffer.
        """
        del self._entries[:]

    def add_entry(
            self,
            module_name: str,
            trace_point: Optional[TracePoint],
            log_type: LogType,
            log_string: str
        ) -> bool:
        """
        Add an entry to the log buffer.

        :param module_name: Name of the module.
        :type module_name: str
        :param trace_point: A TracePoint object representing the
            trace point name to use.
        :type trace_point: Optional[TracePoint]
        :param log_type: The type of log to use.
        :type log_type: LogType
        :param log_string: The string to log.
        :type log_string: str
        :return: ``True`` if the entry was successfully added. ``False``
            otherwise.
        :rtype: bool
        """
        logged = False
        if trace_point is None:
            trace_point_name = ''
        else:
            trace_point_name = trace_point.name
        if len(self._entries) < self._capacity:
            entry = LogEntry(
                module_name,
                datetime.datetime.utcnow(),
                trace_point_name,
                log_string,
                log_type)
            self._entries.append(entry)
            logged = True
        return logged

    @property
    def entries(self) -> List[LogEntry]:
        """
        Get the list of entries in this log buffer.

        :return: The list of entries in this log buffer.
        :rtype: List[LogEntry]
        """
        return self._entries
