# -*- coding: utf-8 -*-
"""
Implementation of 'MessagePublisherBuilder' class
"""
# Import python libs
from typing import Optional

# Import local libs
from systemlink.messagebus.amqp_connection_manager import AmqpConnectionManager
from systemlink.messagebus.trace_logger import TraceLogger


class MessagePublisherBuilder:
    """
    Message Publisher Builder.

    This is used for the construction of a
    :class:`systemlink.messagebus.message_publisher.MessagePublisher` object.
    """
    def __init__(self, origin: Optional[str] = None) -> None:
        """
        :param origin: The ``Origin`` message property which is an
            identifier specifying where messages are sent from. Used
            for message routing. May be ``None``.
        :type origin: Optional[str]
        """
        self._origin = origin
        self._reply_to = None  # type: Optional[str]
        self._connection_manager = None  # type: Optional[AmqpConnectionManager]
        self._trace_logger = None  # type: Optional[TraceLogger]

    @property
    def origin(self) -> Optional[str]:
        """
        Get the ``Origin`` message property.

        :return: The ``Origin`` message property. May be ``None``.
        :rtype: Optional[str]
        """
        return self._origin

    @origin.setter
    def origin(self, origin: Optional[str]) -> None:
        """
        Set the ``Origin`` message property.

        :param origin: The ``Origin`` message property. May be ``None``.
        :type origin: Optional[str]
        """
        self._origin = origin

    @property
    def reply_to(self) -> Optional[str]:
        """
        Get the ``ReplyTo`` message property.

        :return: The ``ReplyTo`` message property. May be ``None``.
        :rtype: Optional[str]
        """
        return self._reply_to

    @reply_to.setter
    def reply_to(self, reply_to: Optional[str]) -> None:
        """
        Set the ``ReplyTo`` message property.

        :param reply_to: The ``ReplyTo`` message property. May be ``None``.
        :type reply_to: str or None
        """
        self._reply_to = reply_to

    @property
    def connection_manager(self) -> Optional[AmqpConnectionManager]:
        """
        Get the Connection Manager. If this is ``None``, a new
        Connection Manager will be created during the creation
        of :class:`systemlink.messagebus.message_publisher.MessagePublisher`.

        :return: The Connection Manager. May be ``None``.
        :rtype: Optional[AmqpConnectionManager]
        """
        return self._connection_manager

    @connection_manager.setter
    def connection_manager(self, connection_manager: Optional[AmqpConnectionManager]) -> None:
        """
        Set the Connection Manager. If this is ``None``, a new
        Connection Manager will be created during the creation
        of :class:`systemlink.messagebus.message_publisher.MessagePublisher`.

        :param connection_manager: The Connection Manager. May be ``None``.
        :type connection_manager: Optional[AmqpConnectionManager]
        """
        self._connection_manager = connection_manager

    @property
    def trace_logger(self) -> Optional[TraceLogger]:
        """
        Get the Trace Logger.

        :return: The Trace Logger. May be ``None``.
        :rtype: Optional[TraceLogger]
        """
        return self._trace_logger

    @trace_logger.setter
    def trace_logger(self, trace_logger: Optional[TraceLogger]) -> None:
        """
        Set the Trace Logger.

        :param trace_logger: The Trace Logger. May be ``None``.
        :type trace_logger: Optional[TraceLogger]
        """
        self._trace_logger = trace_logger
