# -*- coding: utf-8 -*-
"""
Implementation of 'MessageServiceBuilder' class
"""
# Import python libs
from typing import Optional, Union

# Import local libs
from systemlink.messagebus.amqp_connection_manager import AmqpConnectionManager
from systemlink.messagebus.exceptions import SystemLinkException
from systemlink.messagebus.message_subscriber_builder import MessageSubscriberBuilder
from systemlink.messagebus.trace_logger import TraceLogger

DEFAULT_REQUEST_TIMEOUT = 10


class MessageServiceBuilder:
    """
    Message Service Builder.

    This is used for the construction of a
    :class:`systemlink.messagebus.message_service.MessageService` object.
    """
    def __init__(self, service_name) -> None:
        """
        :param service_name: The name of the message service.
        :type service_name: str
        """
        self._instance_name = None  # type: Optional[str]
        if service_name is None:
            error_info = 'Service name cannot be none.'
            raise SystemLinkException.from_name('Skyline.Exception', info=error_info)
        self._service_name = service_name
        self._request_timeout = DEFAULT_REQUEST_TIMEOUT  # type: Union[int, float]
        self._connection_manager = None  # type: Optional[AmqpConnectionManager]
        self._subscriber_builder = None  # type: Optional[MessageSubscriberBuilder]
        self._trace_logger = None  # type: Optional[TraceLogger]

    @property
    def service_name(self) -> str:
        """
        Get the service name.

        :return: The service name.
        :rtype: str
        """
        return self._service_name

    @service_name.setter
    def service_name(self, service_name: str) -> None:
        """
        Set the service name.

        :param service_name: The service name.
        :type service_name: str
        """
        self._service_name = service_name

    @property
    def instance_name(self) -> Optional[str]:
        """
        Get the instance name.

        :return: The instance name. May be ``None``.
        :rtype: Optional[str]
        """
        return self._instance_name

    @instance_name.setter
    def instance_name(self, instance_name: Optional[str]) -> None:
        """
        Set the instance name.

        :param instance_name: The instance name. May be ``None``.
        :type instance_name: str or None
        """
        self._instance_name = instance_name

    @property
    def trace_logger(self) -> Optional[TraceLogger]:
        """
        Get the Trace Logger.

        :return: The Trace Logger.
        :rtype: Optional[TraceLogger]
        """
        return self._trace_logger

    @trace_logger.setter
    def trace_logger(self, trace_logger: Optional[TraceLogger]) -> None:
        """
        Set the Trace Logger.

        :param trace_logger: The Trace Logger.
        :type trace_logger: Optional[TraceLogger]
        """
        self._trace_logger = trace_logger

    @property
    def request_timeout(self) -> Union[int, float]:
        """
        Get the request timeout (in seconds).

        :return: The request timeout (in seconds).
        :rtype: Union[int, float]
        """
        return self._request_timeout

    @request_timeout.setter
    def request_timeout(self, timeout_seconds: Union[int, float]) -> None:
        """
        Set the request timeout (in seconds).

        :param timeout_seconds: The request timeout (in seconds).
        :type timeout_seconds: Union[int, float]
        """
        self._request_timeout = timeout_seconds

    @property
    def connection_manager(self) -> Optional[AmqpConnectionManager]:
        """
        Get the Connection Manager. If this is ``None``, a new
        Connection Manager will be created during the creation
        of :class:`systemlink.messagebus.message_service.MessageService`.

        :return: The Connection Manager. May be ``None``.
        :rtype: Optional[AmqpConnectionManager]
        """
        return self._connection_manager

    @connection_manager.setter
    def connection_manager(self, connection_manager: Optional[AmqpConnectionManager]) -> None:
        """
        Set the Connection Manager. If this is ``None``, a new
        Connection Manager will be created during the creation
        of :class:`systemlink.messagebus.message_service.MessageService`.

        :param connection_manager: The Connection Manager. May be ``None``.
        :type connection_manager: Optional[AmqpConnectionManager]
        """
        self._connection_manager = connection_manager

    @property
    def subscriber_builder(self) -> Optional[MessageSubscriberBuilder]:
        """
        Get the Message Subscriber Builder. If this is ``None``, a new
        Message Subscriber Builder will be created during the creation
        of :class:`systemlink.messagebus.message_service.MessageService`.

        :return: The Message Subscriber Builder. May be ``None``.
        :rtype: Optional[MessageSubscriberBuilder]
        """
        return self._subscriber_builder

    @subscriber_builder.setter
    def subscriber_builder(self, subscriber_builder: Optional[MessageSubscriberBuilder]) -> None:
        """
        Set the Message Subscriber Builder. If this is ``None``, a new
        Message Subscriber Builder will be created during the creation
        of :class:`systemlink.messagebus.message_service.MessageService`.

        :param subscriber_builder: The Message Subscriber Builder. May
            be ``None``.
        :type subscriber_builder: Optional[MessageSubscriberBuilder]
        """
        self._subscriber_builder = subscriber_builder
