# -*- coding: utf-8 -*-
"""
Collection of path-related functions.
"""
# Import python libs
import codecs
import os.path
import sys

# Import local libs
# pylint: disable=import-error
try:
    from systemlink.messagebus.utils import is_windows, is_linux
except ImportError:
    # Import as relative libs (for relocatability)
    # Add the current directory to the search path.
    sys.path.append(os.path.dirname(os.path.abspath(__file__)))
    try:
        from utils import is_windows, is_linux  # type: ignore
    finally:
        # Remove the extra search path that we added to sys.path
        del sys.path[-1:]
# pylint: enable=import-error

SKYLINE_LOCALHOST_FILE_NAME = 'skyline_localhost.json'
SKYLINE_MASTER_FILE_NAME = 'skyline_master.json'
HTTP_LOCALHOST_FILE_NAME = 'http_localhost.json'
HTTP_MASTER_FILE_NAME = 'http_master.json'

if is_windows():
    # pylint: disable=import-error
    import winreg
    # pylint: enable=import-error

    NI_INSTALLERS_REG_PATH = 'SOFTWARE\\National Instruments\\Common\\Installer'
    NI_INSTALLERS_REG_KEY_APP_DATA = 'NIPUBAPPDATADIR'
    NI_INSTALLERS_REG_KEY_SHARED = 'NISHAREDDIR64'

    def _get_ni_shared_dir() -> str:
        """
        Return the National Instruments Shared Application Directory.
        This looks like: 'C:\\Program Files\\National Instruments\\Shared'

        :return: The National Instruments Shared Application Directory.
        :rtype: str
        """
        with winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                NI_INSTALLERS_REG_PATH,
                0,
                winreg.KEY_READ) as hkey:
            (shared_dir, _) = winreg.QueryValueEx(hkey, NI_INSTALLERS_REG_KEY_SHARED)
            return shared_dir

    def _get_ni_common_appdata_dir() -> str:
        """
        Return the National Instruments Common Application Data Directory.
        This looks like: 'C:\\ProgramData\\National Instruments'

        :return: The National Instruments Common Application Data Directory.
        :rtype: str
        """
        with winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                NI_INSTALLERS_REG_PATH,
                0,
                winreg.KEY_READ) as hkey:
            (appdata_dir, _) = winreg.QueryValueEx(hkey, NI_INSTALLERS_REG_KEY_APP_DATA)
            return appdata_dir


def get_application_directory() -> str:
    """
    Returns the directory where the SystemLink binaries may be found.

    :return: Path to the SystemLink application directory.
    :rtype: str
    """
    if is_windows():
        # Get Program Files directory
        ni_shared_path = _get_ni_shared_dir()
        path = os.path.join(
            ni_shared_path,
            'Skyline'
        )
        return path
    return '/usr/local/bin'


def get_application_data_directory() -> str:
    """
    Returns the directory where the SystemLink data files may be found.

    :return: Path to the SystemLink data file directory.
    :rtype: str
    """
    if is_windows():
        # Get APPDATA directory
        ni_appdata_path = _get_ni_common_appdata_dir()
        path = os.path.join(
            ni_appdata_path,
            'Skyline'
        )
        return path
    return '/var/lib/systemlink'


def get_base_configuration_directory() -> str:
    """
    Returns the base directory where SystemLink configuration is stored.

    :return: Path to the base SystemLink configuration directory.
    :rtype: str
    """
    if is_windows():
        return get_application_data_directory()
    return '/etc/systemlink'


def get_legacy_base_configuration_directory() -> str:
    """
    Returns the base directory where SystemLink configuration is stored.
    This is the legacy path on Linux.

    :return: Path to the base SystemLink configuration directory.
    :rtype: str
    """
    if is_windows():
        return get_application_data_directory()
    return '/etc/natinst/niskyline'


def get_configuration_directory() -> str:
    """
    Returns the directory where the SystemLink configuration files may be found.

    :return: Path to the SystemLink configuration directory.
    :rtype: str
    """
    if is_windows():
        return os.path.join(get_base_configuration_directory(), 'Config')
    return os.path.join(get_base_configuration_directory(), 'config')


def get_log_directory() -> str:
    """
    Returns the directory where the SystemLink log files may be found.

    :return: Path to the SystemLink log directory.
    :rtype: str
    """
    if is_windows():
        return os.path.join(get_application_data_directory(), 'Logs')
    return '/var/log/systemlink'


def get_service_descriptor_install_directory() -> str:  # pylint: disable=invalid-name
    """
    Returns the directory where the service descriptors
    are added and removed during install time.

    :return: Path to the service descriptor install directory.
    :rtype: str
    """
    if is_windows():
        return os.path.join(
            get_base_configuration_directory(),
            'Install', 'Services', 'Descriptors'
        )
    return os.path.join(
        get_base_configuration_directory(),
        'install', 'services', 'descs'
    )


def get_skyline_configurations_directory() -> str:  # pylint: disable=invalid-name
    """
    Return path to the SystemLink AMQP configurations directory.

    :return: Path to SystemLink AMQP configurations directory.
    :rtype: str
    """
    if is_windows():
        return os.path.join(
            get_base_configuration_directory(),
            'SkylineConfigurations',
        )
    return os.path.join(
        get_base_configuration_directory(),
        'amqp_config',
    )


def get_legacy_skyline_configurations_directory() -> str:  # pylint: disable=invalid-name
    """
    Return path to the SystemLink AMQP configurations directory.
    This is the legacy path on Linux.

    :return: Path to SystemLink AMQP configurations directory.
    :rtype: str
    """
    return os.path.join(
        get_legacy_base_configuration_directory(),
        'SkylineConfigurations',
    )


def get_http_configurations_directory() -> str:  # pylint: disable=invalid-name
    """
    Return path to the SystemLink HTTP configurations directory.

    :return: Path to SystemLink HTTP configurations directory.
    :rtype: str
    """
    if is_windows():
        return os.path.join(
            get_base_configuration_directory(),
            'HttpConfigurations',
        )
    return os.path.join(
        get_base_configuration_directory(),
        'http_config',
    )


def get_legacy_http_configurations_directory() -> str:  # pylint: disable=invalid-name
    """
    Return path to the SystemLink HTTP configurations directory.
    This is the legacy path on Linux.

    :return: Path to SystemLink HTTP configurations directory.
    :rtype: str
    """
    return os.path.join(
        get_legacy_base_configuration_directory(),
        'HttpConfigurations',
    )


def get_certificates_directory() -> str:  # pylint: disable=invalid-name
    """
    Return path to the SystemLink certificates directory.

    :return: Path to SystemLink certificates directory.
    :rtype: str
    """
    if is_windows():
        return os.path.join(
            get_base_configuration_directory(),
            'Certificates',
        )
    return os.path.join(
        get_base_configuration_directory(),
        'certs',
    )


def get_legacy_certificates_directory() -> str:  # pylint: disable=invalid-name
    """
    Return path to the SystemLink certificates directory.
    This is the legacy path on Linux.

    :return: Path to SystemLink certificates directory.
    :rtype: str
    """
    return os.path.join(
        get_legacy_base_configuration_directory(),
        'Certificates',
    )


def get_skyline_localhost_file() -> str:
    """
    Return path to the SystemLink Localhost AMQP credentials file.

    This API may be called by NI SystemLink Systems Management in a
    legacy Linux environment. Due to this it needs to be able to
    return both legacy and current style paths. In all cases, the
    client will check for file existence before continuing, and
    re-invoke this API if a previous invocation didn't yield an
    existing file.

    :return: Path to SystemLink Localhost AMQP credentials file.
    :rtype: str
    """
    if is_linux():
        path = os.path.join(
            get_legacy_skyline_configurations_directory(),
            SKYLINE_LOCALHOST_FILE_NAME
        )
        if os.path.isfile(path):
            return path
    path = os.path.join(
        get_skyline_configurations_directory(),
        SKYLINE_LOCALHOST_FILE_NAME
    )
    return path


def get_skyline_master_file() -> str:
    """
    Return path to the SystemLink Master AMQP credentials file on the minion.

    This API may be called by NI SystemLink Systems Management in a
    legacy Linux environment. Due to this it needs to be able to
    return both legacy and current style paths. In all cases, the
    client will check for file existence before continuing, and
    re-invoke this API if a previous invocation didn't yield an
    existing file.

    :return: Path to SystemLink Master AMQP credentials file.
    :rtype: str
    """
    if is_linux():
        path = os.path.join(
            get_legacy_skyline_configurations_directory(),
            SKYLINE_MASTER_FILE_NAME
        )
        if os.path.isfile(path):
            return path
    path = os.path.join(
        get_skyline_configurations_directory(),
        SKYLINE_MASTER_FILE_NAME
    )
    return path


def get_http_localhost_file() -> str:
    """
    Return path to the SystemLink Localhost HTTP credentials file.

    This API may be called by NI SystemLink Systems Management in a
    legacy Linux environment. Due to this it needs to be able to
    return both legacy and current style paths. In all cases, the
    client will check for file existence before continuing, and
    re-invoke this API if a previous invocation didn't yield an
    existing file.

    :return: Path to SystemLink Localhost HTTP credentials file.
    :rtype: str
    """
    if is_linux():
        path = os.path.join(
            get_legacy_http_configurations_directory(),
            HTTP_LOCALHOST_FILE_NAME
        )
        if os.path.isfile(path):
            return path
    path = os.path.join(
        get_http_configurations_directory(),
        HTTP_LOCALHOST_FILE_NAME
    )
    return path


def get_http_master_file() -> str:
    """
    Return path to the SystemLink Master HTTP credentials file on the minion.

    This API may be called by NI SystemLink Systems Management in a
    legacy Linux environment. Due to this it needs to be able to
    return both legacy and current style paths. In all cases, the
    client will check for file existence before continuing, and
    re-invoke this API if a previous invocation didn't yield an
    existing file.

    :return: Path to SystemLink Master HTTP credentials file.
    :rtype: str
    """
    if is_linux():
        path = os.path.join(
            get_legacy_http_configurations_directory(),
            HTTP_MASTER_FILE_NAME
        )
        if os.path.isfile(path):
            return path
    path = os.path.join(
        get_http_configurations_directory(),
        HTTP_MASTER_FILE_NAME
    )
    return path


def load_utf8_file_to_string(filename: str) -> str:
    """
    Read the contents of a utf-8 based file and
    return that data as a string.

    :param filename: The input file path.
    :type filename: str
    :return: Contents of utf-8 based file.
    :rtype: str
    """
    with codecs.open(filename, 'r', encoding='utf-8-sig') as fp_:
        data = fp_.read()
    return data
