# -*- coding: utf-8 -*-
"""
Implementation of 'TracePoint' class
"""
# Import python libs
from typing import Optional


class TracePoint:
    """
    A Trace Point for use with
    :class:`systemlink.messagebus.trace_logger.TraceLogger`.
    """
    def __init__(
            self,
            name: Optional[str],
            is_enabled: bool = False,
            full_name: Optional[str] = None,
            declaring_class: Optional[str] = None
        ) -> None:
        """
        :param name: The name of the Trace Point. ``None`` if no name is used.
        :type name: Optional[str]
        :param is_enabled: ``True`` if the Trace Point is currently enabled.
            ``False`` otherwise.
        :type is_enabled: bool
        :param full_name: The full name of the Trace Point. ``None`` if no
            full name is used.
        :type full_name: Optional[str]
        :param declaring_class: Name of the declaring class. ``None`` if no
            declaring class name is used.
        :type declaring_class: Optional[str]
        """
        self._enabled = is_enabled
        self._name = '' if name is None else name
        self._full_name = '<unset>' if full_name is None else full_name
        self._declaring_class = '<unset>' if declaring_class is None else declaring_class

    @property
    def name(self) -> str:
        """
        Get the name of the Trace Point.

        :return: The name of the Trace Point.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """
        Set the name of the Trace Point.

        :param name: The name of the Trace Point.
        :type name: str
        """
        self._name = name

    @property
    def full_name(self) -> str:
        """
        Get the full name of the Trace Point.

        :return: The full name of the Trace Point.
        :rtype: str
        """
        return self._full_name

    @full_name.setter
    def full_name(self, full_name: str) -> None:
        """
        Set the full name of the Trace Point.

        :param full_name: The full name of the Trace Point.
        :type full_name: str
        """
        self._full_name = full_name

    @property
    def declaring_class(self) -> str:
        """
        Get the name of the declaring class.

        :return: The name of the declaring class.
        :rtype: str
        """
        return self._declaring_class

    @declaring_class.setter
    def declaring_class(self, declaring_class: str) -> None:
        """
        Set the name of the declaring class.

        :param declaring_class: The name of the declaring class.
        :type declaring_class: str
        """
        self._declaring_class = declaring_class

    @property
    def is_enabled(self) -> bool:
        """
        Get whether the Trace Point is enabled.

        :return: ``True`` if the Trace Point is enabled. ``False`` otherwise.
        :rtype: bool
        """
        return self._enabled

    def enable(self) -> None:
        """
        Enable the Trace Point.
        """
        self._enabled = True

    def disable(self) -> None:
        """
        Disable the Trace Point.
        """
        self._enabled = False
