# -*- coding: utf-8 -*-
"""
API to create, update, and query Skyline Tags.
"""
from __future__ import absolute_import

# Import python libs
import datetime
import logging
import warnings

# Import local libs
# pylint: disable=import-error
from systemlink.messagebus.amqp_connection_manager import AmqpConnectionManager
from systemlink.messagebus.exceptions import SystemLinkException
from systemlink.messagebus.message_service import MessageService
from systemlink.messagebus.message_service_builder import MessageServiceBuilder
from . import messages as tag_messages  # pylint: disable=no-name-in-module
from . import error_codes
# pylint: enable=import-error

warnings.warn("This AMQP-based TagClient is deprecated. Use the HTTP-based systemlink.clients.nitag instead.", DeprecationWarning )

# Set up logging
LOGGER = logging.getLogger(__name__)


class TagClient(object):
    """
    Class to publicly access the Tag Client.
    """
    def __init__(self,  # pylint: disable=too-many-arguments
                 message_service=None,
                 service_name='TagClient',
                 config=None,
                 connection_timeout=5,
                 auto_reconnect=True):
        """
        :param message_service: An instance of the message
            service to use or ``None`` to allow this object to create and own
            the message service.
        :type message_service: systemlink.messagebus.message_service.MessageService or None
        :param service_name: If `message_service` is ``None`` and therefore
            this object creates and owns the message service, the name to
            use for this message service. Try to pick a unique name for
            your client.
        :type service_name: str or None
        :param config: If
            ``message_service`` is ``None``, the configuration
            to use for this message service. If this is ``None``,
            will use the default configuration.
        :type config: systemlink.messagebus.amqp_configuration.AmqpConfiguration or None
        :param connection_timeout: Timeout, in seconds, to use
            when trying to connect to the message broker.
        :type connection_timeout: float or int
        """
        self._closing = False
        self._own_message_service = False
        self._connection_manager = None
        error_codes.register_error_codes()
        if message_service:
            self._message_service = message_service
        else:
            self._connection_manager = AmqpConnectionManager(config=config)
            self._connection_manager.connection_timeout = connection_timeout
            self._connection_manager.auto_reconnect = auto_reconnect
            message_service_builder = MessageServiceBuilder(service_name)
            message_service_builder.connection_manager = self._connection_manager
            self._message_service = MessageService(message_service_builder)
            self._own_message_service = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Close the TagClient and all associated resources.
        """
        if self._closing:
            return
        self._closing = True
        if self._own_message_service:
            self._message_service.close()
            self._connection_manager.close()

    def create_tag(self, tag_path, tag_type, tag_properties=None,  # pylint: disable=too-many-arguments
                   tag_keywords=None, tag_collect_aggregates=None):
        """
        Create a tag.

        :param tag_path: Identifier for the tag which is typically a dot-separated path.
        :type tag_path: str
        :param tag_type: The data type of the tag.
        :type tag_type: systemlink.tagclient.messages.DataType or str
        :param tag_properties: A dictionary of key/value pairs that
            represent properties of the tag.
        :type tag_properties: dict(str, str) or None
        :param tag_keywords: A list of keywords associated with the tag.
        :type tag_keywords: list(str) or None
        :param tag_collect_aggregates: Whether or not to collect aggregates.
        :type tag_collect_aggregates: bool or None
        """
        if isinstance(tag_type, str):
            tag_type = tag_messages.DataType.from_string(tag_type.upper())
        if tag_collect_aggregates is None:
            if tag_type.value == tag_messages.DataType.STRING:
                tag_collect_aggregates = False
            else:
                tag_collect_aggregates = True

        new_tag = tag_messages.Tag(
            tag_path, tag_type, tag_properties, tag_keywords, tag_collect_aggregates
        )
        msg = tag_messages.TagInfoCreateTagRoutedMessage(new_tag)

        self._message_service.publish_routed_message(msg)

    def update_tag(self, tag_path, tag_type, tag_value, timestamp=None):
        """
        Update a tag.

        :param tag_path: Identifier for the tag which is typically a dot-separated path.
        :type tag_path: str
        :param tag_type: The data type of the tag.
        :type tag_type: systemlink.tagclient.messages.DataType
        :param tag_value: The value of the tag.
        :type tag_value: str or int or float
        :param timestamp: timestamp
        :type timestamp: datetime
        """
        if isinstance(tag_type, str):
            tag_type = tag_messages.DataType.from_string(tag_type.upper())
        if not isinstance(tag_value, str):
            tag_value = str(tag_value)
        typed_value = tag_messages.TypedValue(tag_type, tag_value)
        if not timestamp:
            timestamp = datetime.datetime.utcnow()
        timestamped_value = tag_messages.TimestampedValue(timestamp, typed_value)
        tag_updates = []
        tag_updates.append(timestamped_value)
        update = tag_messages.Update(tag_path, tag_updates)
        updates = []
        updates.append(update)
        LOGGER.debug('Updates = %s', updates)

        routed_message = tag_messages.TagUpdateValuesRoutedMessage(updates)
        self._message_service.publish_routed_message(routed_message)

    def update_tags(self, tag_info):
        """
        Update one or more tags

        :param tag_info: A list of dicts. Each dict must have
            the following keys:
                path (`str`):
                    Identifier for the tag which is typically a
                    dot-separated path.
                type (:class:`systemlink.tagclient.messages.DataType` or `str`):
                    The data type of the tag.
                value (`str` or `int` or `float`):
                    The value of the tag.
        :type tag_info: list(dict)
        """
        updates = []
        timestamp = datetime.datetime.utcnow()

        for tag_item in tag_info:
            tag_updates = []
            tag_path = tag_item['path']
            tag_type = tag_item['type']
            tag_value = tag_item['value']
            if isinstance(tag_type, str):
                tag_type = tag_messages.DataType.from_string(tag_type.upper())
            if not isinstance(tag_value, str):
                tag_value = str(tag_value)
            typed_value = tag_messages.TypedValue(tag_type, tag_value)
            timestamped_value = tag_messages.TimestampedValue(timestamp, typed_value)
            tag_updates.append(timestamped_value)
            update = tag_messages.Update(tag_path, tag_updates)
            updates.append(update)
        LOGGER.debug('Updates = %s', updates)

        routed_message = tag_messages.TagUpdateValuesRoutedMessage(updates)
        self._message_service.publish_routed_message(routed_message)

    def update_tag_metadata(self, tag_path, tag_properties=None,  # pylint: disable=too-many-arguments
                            tag_keywords=None, tag_collect_aggregates=None,
                            replace=False):
        """
        Update the metadata of a tag.

        :param tag_path: Identifier for the tag which is typically a dot-separated path.
        :type tag_path: str
        :param tag_properties: A dictionary of key/value pairs that
            represent properties of the tag.
        :type tag_properties: dict(str, str) or None
        :param tag_keywords: A list of keywords associated with the tag.
        :type tag_keywords: list(str) or None
        :param tag_collect_aggregates: Whether or not to collect aggregates.
        :type tag_collect_aggregates: bool or None
        :param replace: When ``True`` will replace all previous values in tag.
            When ``False``, will override existing values and add new values
            but keep values that exist in the tag and were not specified in this
            invocation.
        :type replace: bool
        """
        if tag_collect_aggregates is None:
            tag_collect_aggregates = tag_messages.Aggregation.from_string('NONE')
        elif tag_collect_aggregates:
            tag_collect_aggregates = tag_messages.Aggregation.from_string('ENABLE')
        else:
            tag_collect_aggregates = tag_messages.Aggregation.from_string('DISABLE')
        routed_message = tag_messages.TagInfoUpdateTagRoutedMessage(
            tag_path,
            tag_keywords,
            tag_properties,
            tag_collect_aggregates,
            replace)
        self._message_service.publish_routed_message(routed_message)

    def query_tags(self, tag_paths):
        """
        Return metadata for one of more tags.

        :param tag_paths: A list of tag identifiers, each
             identifier is typically a dot-separated path.
        :type tag_paths: list(str)
        :return: Metadata for the specified tag(s).
        :rtype: list(systemlink.tagclient.messages.Tag)
        """
        request = tag_messages.TagInfoQueryTagsRequest(tag_paths, {}, [], 0, 0, False)
        generic_message = self._message_service.publish_synchronous_message(request)
        response = self._generic_to_specific_message(
            tag_messages.TagInfoQueryTagsResponse.from_message, generic_message, False
        )
        LOGGER.debug('TotalCount: %d', response.total_count)
        return response.tags

    def query_current_values(self, tag_paths):
        """
        Return current values for one of more tags.

        :param tag_paths: A list of tag identifiers, each
             identifier is typically a dot-separated path.
        :type tag_paths: list(str)
        :return: Current values for the specified tag(s).
        :rtype: list(systemlink.tagclient.messages.CurrentValues)
        """
        request = tag_messages.TagCvQueryCurrentValuesByPathRequest(tag_paths, False)
        generic_message = self._message_service.publish_synchronous_message(request)
        response = self._generic_to_specific_message(
            tag_messages.TagCvQueryCurrentValuesByPathResponse.from_message, generic_message, False
        )
        return response.values

    @staticmethod
    def _generic_to_specific_message(conversion_function, generic_message, allow_error):
        """
        Convert a generic message to a specific message. The message
        type of the specific message is determined by the
        ``conversion_function`` argument.

        :param conversion_function: A callable that takes the generic
            message as the only input argument and returns an instance
            of a specific message type (which is a derived class of
            :class:`systemlink.messagebus.message_base.MessageBase`).
        :type conversion_function: callable
        :param generic_message: The generic message to convert from.
        :type generic_message:
            systemlink.messagebus.generic_message.GenericMessage or
            None
        :param allow_error: If ``True``, will still attempt to
            deserialize to a specific message object even if an error
            is set in the generic message. If ``False``, will always
            raise a :class:`SystemLinkException` when an error is set
            in the generic message.
        :type allow_error: bool
        :return: An instance of the specific message type that is the
            output of the ``convertsion_function`` callable. This is
            a derived class of
            :class:`systemlink.messagebus.message_base.MessageBase`.
        :rtype: systemlink.messagebus.message_base.MessageBase
        :raises SystemLinkException: if a timeout occurred (the
            ``generic_message`` argument is ``None``) or if there is
            an error specified in the generic message and the
            generic message is not convertible via the
            ``conversion_function`` argument or ``allow_error`` is
            ``False``.
        """
        if generic_message is None:
            raise SystemLinkException.from_name('Skyline.RequestTimedOut')
        if generic_message.has_error():
            if not allow_error or not generic_message.body_bytes:
                raise SystemLinkException(error=generic_message.error)
            try:
                return conversion_function(generic_message)
            except Exception:  # pylint: disable=broad-except
                raise SystemLinkException(error=generic_message.error)
        return conversion_function(generic_message)
