# -*- coding: utf-8 -*-
"""
Command line tests for Tag Client API.
"""
from __future__ import absolute_import, print_function

# Import python libs
import argparse
import logging
import sys

# Import local libs
# pylint: disable=import-error
from systemlink.messagebus import utils
from systemlink.tagclient import TagClient
# pylint: enable=import-error


LOG_LEVELS = {
    'all': logging.NOTSET,
    'debug': logging.DEBUG,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
    'info': logging.INFO,
    'warning': logging.WARNING
}

SORTED_LEVEL_NAMES = [
    log_item[0] for log_item in sorted(LOG_LEVELS.items(), key=lambda x: x[1])
]

DEFAULT_LOG_LEVEL = 'warning'

COMMAND_CHOICES = [
    'create', 'update', 'query-tags', 'query-values'
]


def parse_args():
    """
    This will parse the command-line arguments passed in.
    Returns a Namespace object of the parsed command-line arguments.

    :return: A Namespace object of the parsed command-line arguments.
    :rtype: argparse.Namespace
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-l', '--log-level',
        dest='log_level',
        choices=SORTED_LEVEL_NAMES,
        default=DEFAULT_LOG_LEVEL,
        help='Console logging log level. One of {0}. Default: \'{1}\'.'.format(
            ', '.join([repr(l) for l in SORTED_LEVEL_NAMES]),
            DEFAULT_LOG_LEVEL)
    )
    parser.add_argument(
        'command',
        choices=COMMAND_CHOICES,
        help='Run a tag command. One of {0}.'.format(
            ', '.join([repr(cmd) for cmd in COMMAND_CHOICES]))
    )
    parser.add_argument(
        'command-args',
        nargs=argparse.REMAINDER,
        help='Arguments for a specific command.'
    )
    args = parser.parse_args()
    return args


def setup_logging(args):
    """
    This will set up logging at the desired log level.

    :param args: A Namespace object of the parsed command-line arguments.
    :type args: argparse.Namespace
    """
    log_level = LOG_LEVELS[args.log_level]
    log = logging.getLogger()
    log.setLevel(log_level)

    # Create console handler and set level to debug
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)

    # Create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # Add formatter to stream_handler
    stream_handler.setFormatter(formatter)

    # Add handler to logger
    log.addHandler(stream_handler)


def main():  # pylint: disable=too-many-branches
    """
    This is the main entry point for this script.
    """
    args = parse_args()
    setup_logging(args)

    command = args.command
    command_args = getattr(args, 'command-args', [])

    with TagClient(service_name='TagTestClient') as tag_client:
        # parse_args() will ensure that the command is a valid option.
        if command == 'create':
            if not command_args:
                utils.eprint('tag path not specified.')
                sys.exit(1)
            if len(command_args) < 2:
                utils.eprint('tag type not specified')
                sys.exit(1)

            tag_path, tag_type = command_args
            tag_client.create_tag(tag_path, tag_type)
        elif command == 'update':
            if not command_args:
                utils.eprint('tag path not specified.')
                sys.exit(1)
            if len(command_args) < 2:
                utils.eprint('tag type not specified')
                sys.exit(1)
            if len(command_args) < 3:
                utils.eprint('tag value not specified')
                sys.exit(1)

            tag_path, tag_type, tag_value = command_args
            tag_client.update_tag(tag_path, tag_type, tag_value)
        elif command == 'query-tags':
            if not command_args:
                utils.eprint('tag path not specified.')
                sys.exit(1)

            tag_paths = command_args
            tags = tag_client.query_tags(tag_paths)
            for tag in tags:  # pylint: disable=not-an-iterable
                print('path: {0}'.format(tag.path))
                print('\ttype: {0}'.format(tag.type.to_string()))
                print('\tproperties: {0}'.format(tag.properties))
                print('\tkeywords: {0}'.format(tag.keywords))
                print('\tcollect_aggregates: {0}'.format(tag.collect_aggregates))
        elif command == 'query-values':
            if not command_args:
                utils.eprint('tag path not specified.')
                sys.exit(1)

            tag_paths = command_args
            values = tag_client.query_current_values(tag_paths)
            for value in values:  # pylint: disable=not-an-iterable
                print('path: {0}'.format(value.path))
                print('\tcurrent: {0}'.format(value.current.to_dict()))
                if value.aggregates is not None:
                    print('\taggregates: {0}'.format(value.aggregates.to_dict()))

    print('Done.')


if __name__ == '__main__':
    main()
